;  Copyright (c) 1993 Matt Kaufmann and Computational Logic, Inc.
;  (CLI).  All rights reserved.

;; Documentation of @i[proof regions]:

;; A @i[start] is a carriage return followed by the Lisp prompt (see
;; below) or a Pc-Nqthm prompt.  The @i[position associated with a
;; start] is the beginning of the line immediately following the
;; start, except that if there is no such line then it's the end of
;; the line of the start.

;; (By default, the Lisp prompt is `>' followed by any character
;; besides `>', so that interrupts are ignored.  However, this can be
;; reset by running the emacs command @t[checkpoint-options].)

;; The @i[previous start] is the last start that ends before or on the
;; line of the current cursor, or else the buffer's start if there is no
;; such start.  The @i[next start] is the first start that begins after
;; the current cursor (and thus on a later line), or else the buffer's
;; end if there is no such start.  Notice that these notions only
;; depend on the line of the current cursor, not on where the cursor
;; appears on that line.

;; The @i[current proof region] extends from the position associated with
;; the previous start (inclusively) to the position associated with the
;; next start (exclusively).  Notice that these regions are
;; non-overlapping.  Also notice that the line containing a prompt
;; belongs to the preceding proof region, not the next one.  This seems
;; to be the simplest way to deal with the @i[regions] issue if we want to
;; allow the search to start from the last line in the buffer when there
;; is a prompt on that line.  We could check for that, of course, but
;; then if we type a single carriage return this would change things,
;; which is probably bad.

(defvar proved-checkpoints-enabled t)
(defvar next-start-for-maximum-checkpoint-level)
(defvar next-start-for-proved-checkpoints)

(defvar proved-checkpoints)

(defvar checkpoint-level 0)

(defvar maximum-checkpoint-level)

(defvar checkpoint-priority 0)

;; *** Fix the following if we add more priorities.
(defvar max-priority 4)

(defvar checkpoint-line-position t)

(defvar start-over-level)

;; I used to use
;; \\*\\*\\*\\*\\* Now entering the theorem prover \\*\\*\\*\\*\\*:
;; but maybe that's not a good marker really when searching backwards.
;; Maybe it's OK, but it seems that using prompts and control-l as
;; the only `separators' is simpler.

(defvar lisp-prompt ">[^>]")

(defvar find-marker-string-after-prompt
  "\\|
->:\\|
\C-l")

(defvar find-marker-string
  (format "\n%s%s" lisp-prompt find-marker-string-after-prompt))

(defun set-find-marker-string (prompt-string)
  (progn (setq lisp-prompt prompt-string)
	 (setq find-marker-string
	       (format "\n%s%s" lisp-prompt find-marker-string-after-prompt))))

(defun find-previous-marker ()
  "Puts us just before the start of line with the previous marker."
  (re-search-backward find-marker-string nil t))

(defun find-next-marker ()
  "Puts on just after the next marker, hence on a strictly later line."
  (re-search-forward find-marker-string nil t))

(defun previous-start ()
  "Finds the previous start from the end of the current line."
  (let* ((saved-point (point))
	 (success (progn (end-of-line)
			 (find-previous-marker)))
	 (ans (progn (if success
			 (progn (forward-line 2)
				(beginning-of-line)))
		     (point))))
    (goto-char saved-point)
    (if success
	ans
      (point-min))))

(defun next-start ()
  "Finds the next start lying on a strictly later line."
  (let* ((saved-point (point))
	 (success (find-next-marker))
	 (ans (progn (backward-char)	;in case we've already jumped to the next line
		     (end-of-line)
		     (forward-line 1)
		     (beginning-of-line)
		     (point))))
    (goto-char saved-point)
    (if success
	ans
      (point-max))))

(defun go-to-beginning-of-proof ()
  (goto-char (previous-start)))

(defun cp-member (x y)
  ;; from doctor-member
  "Like memq, but uses  equal  for comparison"
  (while (and y (not (equal x (car y))))
    (setq y (cdr y)))
  y)

(defun proved-checkpoints ()
  "Returns the list of all IDs of proved checkpoints."
  (and proved-checkpoints-enabled
       (let ((next-start (next-start)))
	 (if (equal next-start-for-proved-checkpoints next-start)
	     ;; then presumably there's no change in proof output since
	     ;; the last call of this function, so we return that same answer
	     proved-checkpoints
	   (let ((saved-point (point)))
	     (setq proved-checkpoints nil)
	     (setq next-start-for-proved-checkpoints next-start)
	     (go-to-beginning-of-proof)
	     (while (search-forward "!!NOTE PROVED CLAUSES " next-start t)
	       (setq proved-checkpoints
		     (append (read (current-buffer))
			     proved-checkpoints)))
	     (goto-char saved-point)
	     proved-checkpoints)))))

(defun next-checkpoint ()
  "Part of the Nqthm checkpoint facility, this puts the cursor just above
the checkpoint of next-lower significance relative to the checkpoint
most recently visited."
  (interactive)
  (if (not (and (boundp 'maximum-checkpoint-level)
		maximum-checkpoint-level))
      (error "You must go to the first checkpoint before asking for the next.")
    (let ((current-checkpoint-level checkpoint-level)
	  (current-checkpoint-priority checkpoint-priority))
      (catch 'next-checkpoint
	(while (<= checkpoint-level (maximum-checkpoint-level))
	  (increment-checkpoint-level-and-priority)
	  (let ((ans (next-checkpoint-at-current-level)))
	    (if ans
		(throw 'next-checkpoint ans))))
	(progn (setq checkpoint-level current-checkpoint-level)
	       (setq checkpoint-priority current-checkpoint-priority)
	       (beep)
	       (message "No more checkpoints.")
	       nil)))))

(defun next-checkpoint-at-current-level ()
  (catch 'next-checkpoint-at-current-level-exit
    (progn (while (not (find-one-checkpoint))
	     (if (= checkpoint-priority max-priority)
		 (throw 'next-checkpoint-at-current-level-exit nil)
	       (setq checkpoint-priority (+ 1 checkpoint-priority))))
	   t)))

(defun increment-checkpoint-level-and-priority ()
  (if (= checkpoint-priority max-priority)
      (progn (setq checkpoint-priority 0)
	     (setq checkpoint-level (+ 1 checkpoint-level)))
    (setq checkpoint-priority (+ 1 checkpoint-priority))))

(defun previous-checkpoint ()
  "Part of the Nqthm checkpoint facility, this puts the cursor just above
the checkpoint of next-higher significance relative to the checkpoint
most recently visited."
  (interactive)
  (if (not (and (boundp 'maximum-checkpoint-level)
		maximum-checkpoint-level))
      (error "You must go to the first checkpoint before asking for the previous.")
    (let ((current-checkpoint-level checkpoint-level)
	  (current-checkpoint-priority checkpoint-priority))
      (catch 'previous-checkpoint
	(while (>= checkpoint-level 0)
	  (decrement-checkpoint-level-and-priority)
	  (let ((ans (previous-checkpoint-at-current-level)))
	    (if ans
		(throw 'previous-checkpoint ans))))
	(progn (setq checkpoint-level current-checkpoint-level)
	       (setq checkpoint-priority current-checkpoint-priority)
	       (beep)
	       (message "No preceding checkpoints.")
	       nil)))))

(defun previous-checkpoint-at-current-level ()
  (catch 'previous-checkpoint-at-current-level-exit
    (progn (while (not (find-one-checkpoint))
	     (if (= checkpoint-priority 0)
		 (throw 'previous-checkpoint-at-current-level-exit nil)
	       (setq checkpoint-priority (- checkpoint-priority 1))))
	   t)))

(defun decrement-checkpoint-level-and-priority ()
  (if (= checkpoint-priority 0)
      (progn (setq checkpoint-priority max-priority)
	     (setq checkpoint-level (- checkpoint-level 1)))
    (setq checkpoint-priority (- checkpoint-priority 1))))

(defun find-one-checkpoint ()
  (interactive)
  (let ((saved-point (point))
	(done nil))
    (go-to-beginning-of-proof)
    (let ((next-start (next-start)))
      (if (progn
	    (while (and (not done)
			(search-forward 
			 (format "!!CHECKPOINT LEVEL %s; PRIORITY %s; ID "
				 checkpoint-level checkpoint-priority)
			 next-start t))
	      (setq done (not (memq (read (current-buffer)) (proved-checkpoints)))))
	    done)
	  (progn (push-mark saved-point)
		 (backward-char)
		 (beginning-of-line)
		 (if (or (eq checkpoint-line-position nil)
			 (numberp checkpoint-line-position))
		     (recenter checkpoint-line-position))
		 t)
	(progn (goto-char saved-point)
	       nil)))))

(defun maximum-checkpoint-level ()
  ;; returns nil if there are no checkpoints
  (let ((saved-point1 (point)))
    (forward-line -1)
    (let ((next-start (next-start)))
      (if (and (equal next-start-for-maximum-checkpoint-level next-start)
	       maximum-checkpoint-level)
	  ;; ... then we have a good guess; let's go with it.
	  (progn (goto-char saved-point1)
		 maximum-checkpoint-level)
	(let ((ans nil))
	  (setq next-start-for-maximum-checkpoint-level next-start)
	  (goto-char next-start)
	  ;; we may be a line too late, so let's be sure not simply
	  ;; to find the `start' we're already looking at
	  (forward-line -1)
	  (beginning-of-line)
	  (let ((new-point (point)))
	    (let ((marker (progn (find-previous-marker)
				 (point))))
	      (goto-char new-point)
	      (if (and marker (search-backward "!!CHECKPOINT LEVEL" marker t))
		  (progn (forward-char 19)
			 (setq ans (read (current-buffer)))))
	      (goto-char saved-point1)
	      (setq maximum-checkpoint-level ans))))))))

(defun look-for-start-over ()
  (let ((saved-point (point))
	(next-start (next-start)))
    (go-to-beginning-of-proof)
    (setq start-over-level nil)
    (if (search-forward "disregard" next-start t)
	(let ((next-level (if (search-forward "!!CHECKPOINT LEVEL " next-start t)
			      (read (current-buffer))
			    nil)))
	  (beep)
	  (setq start-over-level next-level)
	  (if next-level
	      (message
	       (format "Note: the prover is inducting on the input, starting at level %s. Try C-t g."
			next-level))
	    (message (format "Note: the prover is inducting on the input."
			     next-level)))))
    (goto-char saved-point)))

(defun first-checkpoint ()
  "Part of the Nqthm checkpoint facility, this starts the search for checkpoints
in a failed (or failing) Nqthm proof.  It puts the cursor just above the most
significant checkpoint."
  (interactive)
  (setq checkpoint-level -1)
  (setq checkpoint-priority max-priority)
  (setq next-start-for-proved-checkpoints nil)
  (setq next-start-for-maximum-checkpoint-level nil)
  (setq maximum-checkpoint-level
	(maximum-checkpoint-level))
  (if maximum-checkpoint-level
      (progn (next-checkpoint)
	     (look-for-start-over))
    (error "No checkpoints appear in this proof")))

(defun checkpoint-options ()
  "Used interactively to set options for the Nqthm checkpointing feature."
  (interactive)
  (if (y-or-n-p
       (format "Toggle ignoring of proved checkpoints (currently %s)? "
	       (if proved-checkpoints-enabled 'on 'off)))
      (setq proved-checkpoints-enabled
	    (not proved-checkpoints-enabled)))
  (if (y-or-n-p "Change what is considered to be the Lisp prompt? ")
      (set-find-marker-string (read-from-minibuffer "Enter prompt: ")))
  (if (or (eq checkpoint-line-position t)
	  (eq checkpoint-line-position nil))
      (if (y-or-n-p "Set cursor line number for checkpoints? ")
	  (setq checkpoint-line-position (read-minibuffer "Enter line number: "))
	(if (eq checkpoint-line-position t)
	    (if (y-or-n-p "Begin recentering cursor line for checkpoints? ")
		(setq checkpoint-line-position nil))
	  (if (y-or-n-p "Stop recentering cursor line for checkpoints? ")
	      (setq checkpoint-line-position t))))
    (if (y-or-n-p "Stop setting cursor line number for checkpoints? ")
	(if (y-or-n-p "Begin recentering cursor line for checkpoints? ")
	    (setq checkpoint-line-position nil)
	  (setq checkpoint-line-position t))))
  (message "Done. For a related utility, submit (CHECKPOINT-OPTIONS) to Nqthm."))

(defun goto-checkpoint-level ()
  "Go to the ``best'' checkpoint at the indicated level."
  (interactive)
  (let ((n (read-from-minibuffer "Go to checkpoint at level: "
				 (if start-over-level
				     (format "%s" start-over-level)
				   nil)
				 nil t)))
    (let ((saved-priority checkpoint-priority)
	  (saved-level checkpoint-level))
      (setq checkpoint-priority 0)
      (setq checkpoint-level n)
      (if (next-checkpoint-at-current-level)
	  (progn (setq start-over-level nil)
		 t)
	(progn (beep)
	       (message (format "There is no checkpoint at level %s." n))
	       (setq checkpoint-priority saved-priority)
	       (setq checkpoint-level saved-level))))))
