;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

;;; Adapted from Boyer and Moore's nqthm.lisp.

;  This file pc-nqthm.lisp is the basic loader and compiler for the
;  Pc-Nqthm interactive enhancement of the Boyer-Moore Theorem Prover.
;  To compile the system, after loading this file into a Common Lisp
;  with Nqthm-1992 loaded, invoke (IN-PACKAGE "USER") and then
;  (COMPILE-PC-NQTHM).  To build the system for execution, assuming
;  that it has been compiled and after loading this file, invoke
;  (IN-PACKAGE "USER") and then (LOAD-PC-NQTHM).  One can then proceed
;  to use BOOT-STRAP, PROVE-LEMMA, and the other commands described in
;  the Nqthm manual.  For further guidance in case of problems, see the
;  installation guide chapter of the manual for Nqthm-1992.

;  When installing this system, before invoking (COMPILE-PC-NQTHM) or
;  (LOAD-PC-NQTHM), it may be desirable to set *DEFAULT-NQTHM-PATH* in
;  order to specify the directory on which the sources are located and
;  to which the objects are to be written.  For example, one might
;  wish to (SETQ *DEFAULT-NQTHM-PATH* "/usr/my/pc-nqthm/"), after
;  invoking (IN-PACKAGE "USER").  For some implementations of Common
;  Lisp the default setting of NIL will cause the "current" or
;  "connected" directory to be used, which is ok provided that the
;  connected directory is the one with the sources and provided your
;  operating system supports, as does Unix, the concept of a
;  `connected' directory for a process.

;;;  As with Nqthm, and as explained in the Nqthm file nqthm.lisp,
;;;  we do everything in the USER package.  We assume that the
;;;  finding and/or creation of the USER and SLOOP packages has
;;;  already been done by nqthm.lisp.

(IN-PACKAGE "USER")

;;; I assume that the CHK-BASE-AND-PACKAGE-1992 check has already been
;;; done by Nqthm.

;;; Recall from Nqthm's file nqthm.lisp that *DEFAULT-NQTHM-PATH* may
;;; be set by the user to ``set the directory from which all files are
;;; to be loaded and to which files are to be printed.  The user is
;;; free to set this variable.''  It should be a string, e.g.
;;; "/usr/me/pc-nqthm", or NIL.

;;; I assume that FILE-EXTENSION-LISP and FILE-EXTENSION-BIN have
;;; already been set by Nqthm.

;;; First let us see if this Lisp has the kind of support we need for
;;; redefinition of functions.  For example, kcl (as opposed to akcl)
;;; may not allow such redefinition during the kind of ``block
;;; compilation'' that is done in building Pc-Nqthm-1992.  Perhaps the
;;; forms below should be part of the definition of COMPILE-PC-NQTHM,
;;; but I don't want to do anything ``funny'' here, and I find it
;;; ``funny'' to embed a DEFUN inside a DEFUN in this case.  The idea
;;; here came from a suggestion of Bob Boyer, to whom I am grateful.

(compile-file (extend-file-name "lisp-test" file-extension-lisp))
(load (extend-file-name "lisp-test" file-extension-bin))
(defun lisp-test-foo2 () 3)
(when (equal (lisp-test-foo1) 2)
      #+kcl
      (princ "
**** You appear to be using kcl, but it is strongly recommended
that you use akcl instead."
	     t)
      (error "
This Lisp probably won't work with Pc-Nqthm.  The problem is that your
Lisp supports a style of compilation that can compile a function, say
FOO-1, that calls another function, say FOO-2, in such a manner that
even after we redefine FOO-2, FOO-1 continues to use the old
definition of FOO-2.
"))

;  Invoking (COMPILE-PC-NQTHM) is all it takes to compile this
;  system.  The order of compilation and loading is
;  significant; files may define macros that are used in later files.
(DEFUN COMPILE-PC-NQTHM ()
  (FLET ((LF (N)
             (LOAD (EXTEND-FILE-NAME N FILE-EXTENSION-BIN)))
         (CF (N)
             (COMPILE-FILE (EXTEND-FILE-NAME N FILE-EXTENSION-LISP))))
        #+akcl ;keep this in sync with directive in Nqthm's proclaim-nqthm-file
               ;and make-declare-form
        (LOAD (EXTEND-FILE-NAME "proclaim-nqthm" FILE-EXTENSION-LISP))
	(PROCLAIM-PC-FILES)
	(CF "defn-sk")
	(CF "nqthm-patches")
	(LF "nqthm-patches")
	(CF "basis-pc")
	(LF "basis-pc")
	(CF "help")
	(CF "utilities")
	(CF "top")
	(CF "top-nqthm")
	(CF "macro-commands-aux")
	(LF "macro-commands-aux")
	(CF "macro-commands")))

;  Invoking (LOAD-PC-NQTHM) is all it takes to build a runnable version of
;  this system, assuming that you have compiled it.
(DEFUN LOAD-PC-NQTHM (&OPTIONAL (BOOT-STRAP-FLG 'NQTHM))
  (FLET ((LF (N)
             (LOAD (EXTEND-FILE-NAME N FILE-EXTENSION-BIN))))
	(LF "defn-sk")
	(LF "nqthm-patches")
	(LF "basis-pc")
	(LF "help")
	(LF "utilities")
	(LF "top")
	(LF "top-nqthm")
	(LF "macro-commands-aux")
	(LF "macro-commands"))
  ;; Finally, call BOOT-STRAP in this environment, to initialize the modified prover.
  ;; It's OK to replace nqthm by thm here.
  (eval `(boot-strap ,boot-strap-flg)))

;;; Nothing from the macro commands files should be necessary for replay.
(DEFUN LOAD-PC-NQTHM-SHORT (&OPTIONAL (BOOT-STRAP-FLG 'NQTHM))
  (FLET ((LF (N)
             (LOAD (EXTEND-FILE-NAME N FILE-EXTENSION-BIN))))
	(LF "defn-sk")
	(LF "nqthm-patches")
	(LF "basis-pc")
	(LF "help")
	(LF "utilities")
	(LF "top")
	(LF "top-nqthm"))
  ;; Finally, call BOOT-STRAP in this environment, to initialize the modified prover.
  ;; It's OK to replace nqthm by thm here.
  (eval `(boot-strap ,boot-strap-flg)))

;;; I assume that *NQTHM-MAKE-PROCLAMATIONS* has already been set
;;; appropriately by Nqthm.

(DEFUN PROCLAIM-PC-FILES ()
  (DOLIST
      (NAME '("defn-sk" "nqthm-patches" "basis-pc" "help" "utilities" "top"
	      "top-nqthm" "macro-commands-aux"))
    (PROCLAIM-NQTHM-FILE NAME)))
