;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

; This file contains the utilities called by the state-changing functions.

(DEFUN ABBREVIATE (TERM ABBREVIATIONS)
  
  ;; This function assumes that TERM is a term in translated form and that
  ;; ABBREVIATIONS is a list of pairs whose CDRs are to be substituted by
  ;; its CARs (which are presumably variables beginning with @), from the
  ;; outside in.  I use SUBLIS-EXPR1 instead of SUBLIS-EXPR because it is
  ;; perfectly OK to abbreviate a quoted expression -- well actually,
  ;; SUBLIS-EXPR doesn't exist in Nqthm-1992 anyhow!
  
  (if (null abbreviations)
      term
    (SUBLIS-EXPR1 (ITERATE FOR PAIR IN ABBREVIATIONS
			   COLLECT (CONS (CDR PAIR) (CAR PAIR)))
		  TERM)))

(DEFUN ALL-ALMOST-PRIMITIVE-REWRITE-RULES (NEW-LEMMA-NAMES)
  
  ;; Collects all almost primitive fast rewrite rules in the current context,
  ;; i.e., chronology, including the primitive rewrite rules.  The result is
  ;; a function-rules alist.  NOTE:  It's harmless if NEW-LEMMA-NAMES contains
  ;; names that aren't names of lemmas.
  
  (ITERATE FOR X IN CHRONOLOGY
	   NCONC
	   (ITERATE FOR Y IN
		    (CONS X (GET X 'SATELLITES))
		    WHEN (ROSE-FUNCTIONP Y)
		    COLLECT (CONS Y (FREWRITE-RULES-FOR Y NEW-LEMMA-NAMES)))))

(DEFUN ALL-FNNAMES-LST (LST)
  ;; This is taken exactly from Nqthm, but included because this function
  ;; has been deleted in Nqthm-1992.
  (LET (ANS)
    (ITERATE FOR X IN LST DO (ALL-FNNAMES1 X))
    ANS))

(DEFUN ALL-GOALS (STATE)
  (CONS (ACCESS STATE GOAL STATE)
	(ITERATE FOR GL-INFO IN (ACCESS STATE OTHER-GOALS STATE)
	      COLLECT (CAR GL-INFO))))

(defun all-subv-pairs (assumptions)
  ;; I'll reverse the assumptions with the idea being that the most "recent" is the
  ;; one we want to use, in case a variable is equated to two different things.
  (iterate for x in (reverse assumptions)
	   with var and term and vars-so-far
	   when (and (or (and (match x (equal var term)) 
			      (variablep var))
			 (and (match x (equal term var))
			      (variablep var)))
		     ;; the next conjunct is to keep the substitution small rather
		     ;; than for soundness
		     (not (member-eq var vars-so-far))
		     (not (intersectp-eq (cons var vars-so-far) (all-vars term))))
	   collect (progn (setq vars-so-far (append (cons var (all-vars term))
						    vars-so-far))
			  (cons var term))))

(DEFUN ALL-VARS-GOAL (GOAL)
  (UNION-EQ (ALL-VARS (ACCESS GOAL CONC GOAL))
	    (ALL-VARS-LST (MAPCAR #'CAR (ACCESS GOAL HYPS GOAL))))) 

(DEFUN ALL-VARS-STATE (GOAL OTHER-GOALS)
  ;; I contemplated a function FREE-FOR-STATE which would check a variable
  ;; for freeness, but rather than saving CONSes it would cost them, since
  ;; ALL-VARS would be called every time a term was looked at.
  (UNION-EQ (ALL-VARS-GOAL GOAL)
	    (UNION-EQ-LIST
	     (ITERATE FOR GL-INFO IN OTHER-GOALS
		      COLLECT (ALL-VARS-GOAL (CAR GL-INFO))))))

(DEFUN ALMOST-PRIMITIVE-REWRITE-RULEP (LEMMA)
  ;; A rewrite rule is FAST if its conclusion is an equality, every
  ;; variable in a hypothesis occurs in the left side of the conclusion, and the
  ;; rewrite rule is not permutative.  Such rules are also called "almost
  ;; primitive".
  ;; The argument LEMMA has record-type LEMMA.

  (AND
   ;;(MATCH (ACCESS REWRITE-RULE CONCL LEMMA) (EQUAL & &))
   ;;not needed, as this function is called after making equality
   (SUBSETP-EQ (ALL-VARS-LST (ACCESS REWRITE-RULE HYPS LEMMA))
	       (ALL-VARS (FARGN (ACCESS REWRITE-RULE CONCL LEMMA) 1)))
   (NULL (ACCESS REWRITE-RULE LOOP-STOPPER LEMMA))))

(DEFUN APPLICABLE-REWRITES (TERM ID-IFF &OPTIONAL TARGET STATEMENT)
; Collects the lemmas whose conclusions can rewrite TERM under the given
; ID-IFF flag, whether or not the hypotheses are relieved, each consed onto
; the substitution which one-way-unifies the term with the left-hand-side
; of the lemma.  Each lemma is returned in a form (EQUAL LHS RHS).
; However, TARGET is a positive integer or a symbol; if non-NIL, we are only
; interested in those lemmas which "agree" with the target.
;    This function assumes that TERM is not a variable.

;; Patched 8/2/88 to properly handle the case of multiple rewrites with
;; the same name, so that the stored command really works when the
;; MAKE-EQUALITY done by REWRITE-CURRENT-TERM mungs such lemmas.

  (COND ((AND TARGET (SYMBOLP TARGET))
	 ; in this case we ignore shelved lemmas
	 (LET (SHELVED-LEMMAS)
	   (ITERATE FOR LEMMA IN (GET (FN-SYMB TERM) (QUOTE LEMMAS))
		    WITH TEMP
		    WHEN (AND
			  (EQ (ACCESS REWRITE-RULE NAME LEMMA) TARGET)
			  (SETQ TEMP
				(REWRITE-CURRENT-TERM
				 TERM LEMMA ID-IFF))
			  (COND
			   (STATEMENT
			    (EQUAL STATEMENT
				   (MAKE-IMPLICATION
				    (ACCESS REWRITE-RULE HYPS (CAR TEMP))
				    (ACCESS REWRITE-RULE CONCL (CAR TEMP)))))
			   (T T)))
		    COLLECT TEMP)))
	(T
	 (LET ((SHELVED-LEMMAS (AND SHELVED-LEMMAS-FLG SHELVED-LEMMAS)))
	   (IF (INTEGERP TARGET)
	       (ITERATE FOR LEMMA IN (GET (FN-SYMB TERM) (QUOTE LEMMAS))
			WITH TEMP AND I = 1
			DO
			(COND ((SETQ TEMP
				     (REWRITE-CURRENT-TERM
				      TERM LEMMA ID-IFF))
			       (IF (INT= I TARGET)
				   (RETURN (LIST TEMP))
				   (SETQ I (1+ I)))))
			FINALLY (RETURN NIL))
	       (ITERATE FOR LEMMA IN (GET (FN-SYMB TERM) (QUOTE LEMMAS))
			WITH TEMP
			WHEN (SETQ TEMP
				   (REWRITE-CURRENT-TERM
				    TERM LEMMA ID-IFF))
			COLLECT TEMP))))))

(DEFUN ASSERT-TERM (TERM TYPE-ALIST)
; The following function returns an equivalent term under the assumption
; that the term's value is not F.
  (IF (TSLOGSUBSETP (TYPE-SET TERM) TYPE-SET-BOOLEAN)
      TRUE
      (FCONS-TERM* (QUOTE IF) TERM TERM TRUE)))

(DEFUN ASSUMPTIONS (HYP-FLGS)
  (ITERATE FOR HYP-FLG IN HYP-FLGS
	WHEN (AND (EQ (CDR HYP-FLG) (QUOTE A))
		  (NOT (EQUAL (CAR HYP-FLG) TRUE)))
	  COLLECT (CAR HYP-FLG)))

(DEFUN ASSUMPTIONS2 (HYP-FLGS1 HYP-FLGS2)
  ;; same as (NCONC (ASSUMPTIONS HYP-FLGS1) (ASSUMPTIONS HYP-FLGS2)), but
  ;; I don't want to rely on ASSUMPTIONS not sharing structure so I don't
  ;; call ASSUMPTIONS here.  
  (NCONC (ITERATE FOR HYP-FLG IN HYP-FLGS1
		  WHEN (AND (EQ (CDR HYP-FLG) (QUOTE A))
			    (NOT (EQUAL (CAR HYP-FLG) TRUE)))
		  COLLECT (CAR HYP-FLG))
	 (ITERATE FOR HYP-FLG IN HYP-FLGS2
		  WHEN (AND (EQ (CDR HYP-FLG) (QUOTE A))
			    (NOT (EQUAL (CAR HYP-FLG) TRUE)))
		  COLLECT (CAR HYP-FLG))))

(DEFUN BASH (FORM)

  ; Notice that the code is patterned after that for PROVE, except that
  ; we simply return the list of clauses whose provability implies the
  ; provability of FORM rather than returning T or NIL to indicate
  ; something.  We also side-effect PROCESS-CLAUSES similarly as well
  ; as PROCESS-HIST.

  (LET ((PROOF-START-TIME-IN-60THS (TIME-IN-60THS)))
    (PROG1
          (PROG (THM CLAUSES VARS)
                (CHK-INIT)
                (SETQ THM (TRANSLATE-AND-CHK FORM))
                (SETQ CLAUSES (PREPROCESS THM))
                (bash-SETUP FORM ABBREVIATIONS-USED)
                (SETQ VARS (ITERATE FOR CL IN CLAUSES WITH ITERATE-ANS
                                    DO (SETQ ITERATE-ANS
                                             (UNION-EQ (ALL-VARS-LST CL)
                                                       ITERATE-ANS))
                                    FINALLY (RETURN ITERATE-ANS)))
                (SETQ ELIM-VARIABLE-NAMES1 (SET-DIFF ELIM-VARIABLE-NAMES VARS))
                (SETQ GEN-VARIABLE-NAMES1 (SET-DIFF GEN-VARIABLE-NAMES VARS))
                                        ;             (simplify-clause-maximally (car clauses))
                (my-simplify-clauses-maximally clauses) ;sets PROCESS-CLAUSES
                )
      (SETQ PROVE-TIME (- (- (TIME-IN-60THS) PROOF-START-TIME-IN-60THS)
                          IO-TIME)))))

(DEFUN bash1 (TERM)
  ;; we should be sure that HINTS is always bound when this function is called
 (let ((IN-PROVE-LEMMA-FLG T))
  (BLURB NIL CR |*****| |Now| |entering| |the| |theorem|
	 |prover's| |rewriter - simplifier| |*****| CR)
  (UNWIND-PROTECT
   (LET (NEW-PROVE-TIME NEW-IO-TIME PROVE-ANS)
	(LET ((PROVE-TIME 0)
	      (IO-TIME 0))
	     (SETQ PROVE-ANS (bash TERM))
	     (SETQ NEW-PROVE-TIME PROVE-TIME)
	     (SETQ NEW-IO-TIME IO-TIME))
	(SETQ PROVE-TIME (+ PROVE-TIME NEW-PROVE-TIME))
	(SETQ IO-TIME (+ IO-TIME NEW-IO-TIME))
	PROVE-ANS)
   (PROGN (ITERATE FOR X IN HINT-VARIABLE-ALIST
		   DO (SET (CADR X) (CADDDR X)))
	  (SETQ LOCAL-DISABLEDP-HASH-FLAG NIL)))))

(DEFUN bash-SETUP (FORM LEMMAS)
; taken from SETUP
  (SETQ ORIGTHM FORM)
  (SETQ EXPAND-LST HINTED-EXPANSIONS)
  (SETQ TERMS-TO-BE-IGNORED-BY-REWRITE NIL)
  (SETQ INDUCTION-HYP-TERMS NIL)
  (SETQ INDUCTION-CONCL-TERMS NIL)
  (SETQ ALL-LEMMAS-USED LEMMAS)
  (SETQ STACK NIL)
  (SETQ FNSTACK NIL)
  (SETQ LAST-PRINT-CLAUSES NIL)
  (SETQ TYPE-ALIST NIL)
  (SETQ LITS-THAT-MAY-BE-ASSUMED-FALSE NIL)
  (SETQ CURRENT-LIT 0)
  (SETQ CURRENT-ATM 0)
  (SETQ ANCESTORS NIL)
  (COND ((and (boundp 'REWRITE-PATH-STK-PTR)
              REWRITE-PATH-STK-PTR)
         (SETQ REWRITE-PATH-STK-PTR -1)
         (SETQ REWRITE-PATH-FRAME-CNT 0)
         (SETQ REWRITE-PATH-PERSISTENCE-ALIST NIL)))
  (INIT-LEMMA-STACK)
  (INIT-LINEARIZE-ASSUMPTIONS-STACK)
  (SETQ LAST-PRINEVAL-CHAR (QUOTE |.|))
  (RANDOM-INITIALIZATION ORIGTHM)
;  (IO (QUOTE SETUP)
;      (LIST ORIGTHM)
;      NIL CLAUSES LEMMAS)
  )

(DEFUN BEST-SUBST-LEMMA-VARS-HYPS
  (LEMMA-SUBST-LST ASSUMPTIONS DOLLAR-SUBST)
  ;; Returns NIL if there are no "applicable rewrites" in the list
  ;;  LEMMA-SUBST-LST.  Otherwise, returns a version of the
  ;; first applicable rewrite which generates no hypotheses to relieve (if
  ;; any, else simply of the first applicable rewrite in LEMMA-SUBST-LST).
  ;; What is really returned (if not NIL) is an object of the form
  ;;    (SUBST LEMMA VARS . HYPS)
  ;; exactly as in LEMMA-VARS-HYPS-REMAINING, except that SUBST is just
  ;; passed along from the appropriate member of LEMMA-SUBST-LST, and
  ;; dollar signs are removed from the hypotheses.
  ;;   This now returns 'CONTRADICTION if the ASSUMPTIONS are contradictory
  ;; in the sense of type-set reasoning.
  
  (LET ((TYPE-ALIST
	 (LET (TYPE-ALIST)
	      (FLESH-OUT-TYPE-ALIST
	       (MY-TYPE-ALIST-CLAUSE
		(ITERATE FOR X IN ASSUMPTIONS
			 COLLECT (DUMB-NEGATE-LIT X)))))))
       (COND
	((EQ TYPE-ALIST 'CONTRADICTION)
	 'CONTRADICTION)
	(LEMMA-SUBST-LST		;otherwise NIL
	 (SETQ ASSUMPTIONS (FLESH-OUT-ASSUMPTIONS ASSUMPTIONS))
	 (LET ((BACKUP-RESULT
		(CONS (CDAR LEMMA-SUBST-LST)
		      (LEMMA-VARS-HYPS-REMAINING
		       (CAAR LEMMA-SUBST-LST) (CDAR LEMMA-SUBST-LST) DOLLAR-SUBST
		       ASSUMPTIONS TYPE-ALIST))))
	      (IF (NULL (CDDDR BACKUP-RESULT))
		  BACKUP-RESULT
		  (ITERATE FOR LEMMA-SUBST IN (CDR LEMMA-SUBST-LST)
			   WITH TEMP-LEMMA-VARS-HYPS
			   WHEN
			   (NULL (CDDR (SETQ TEMP-LEMMA-VARS-HYPS
					     (LEMMA-VARS-HYPS-REMAINING
					      (CAR LEMMA-SUBST)
					      (CDR LEMMA-SUBST)
					      DOLLAR-SUBST
					      ASSUMPTIONS TYPE-ALIST))))
			   DO
			   (RETURN (CONS (CDR LEMMA-SUBST) TEMP-LEMMA-VARS-HYPS))
			   FINALLY (RETURN BACKUP-RESULT))))))))

(defun boyer-moore-print (fmla leftmargin
			       &optional
			       (rparcnt 0)
			       (pprfile prove-file)
			       (level (or *print-level* -1))
			       (length (or *print-length* -1)))
  ;; takes advantage of *print-level* and *print-length*
  (pprind (fix-level-and-length fmla level length) leftmargin rparcnt pprfile))

(DEFUN CHANGE-TO-UNPROVED-GOAL (STATE)
  
; If the current goal of STATE does not have a conclusion of TRUE,
; then STATE is returned.  Otherwise, a new state is returned which
; results from STATE by switching to a goal which has not yet
; been proved, if there is one; otherwise, STATE is returned.  Useful
; messages are printed along the way.
  
  (LET-FIELDS
    GOAL-NAME DEPENDS-ON CONC	;for the compiler's contentment
    (COND
      ((EQUAL (ACCESS GOAL CONC GOAL) TRUE)
       (LET (GOAL-AND-INFO ALL-GOALS DEP-NAMES NEW-NAME
	     (NAME (ACCESS GOAL GOAL-NAME GOAL)))
	 ; First print information about the change of current goal
	 (COND
	   ((ACCESS GOAL DEPENDS-ON GOAL)
	    (SETQ ALL-GOALS (ALL-GOALS STATE))
	    (SETQ DEP-NAMES
		  (ITERATE FOR GL IN (UNPROVED-DEPENDENTS GOAL ALL-GOALS)
			COLLECT (ACCESS GOAL GOAL-NAME GL)))
	    (COND
	      ((NULL DEP-NAMES)	;the goal and its dependents have been proved
	       (BLURB (NAME) CR |The| |current| |goal| |,| (!PPR NAME (QUOTE |,|))
		      |has| |been| |proved| |,| |as| |have| |all| |of|
		      |its| |dependents| |.|))
	      (T		;otherwise, print the names of the non-T dependents
	       (BLURB (NAME DEP-NAMES)
		      CR |The| |proof| |of| |the| |current| |goal| |,|
		      (!PPR NAME (QUOTE |,|)) |has| |been| |completed| |.| |However|
		      |,| |the| |following| (plural? dep-names |subgoals| |subgoal|)
		      |of| (!PPR NAME NIL) (plural? dep-names |remain| |remains|)
		      |to| |be| |proved| |:| (!PPR-LIST DEP-NAMES (QUOTE |.|)) CR))))
	   (T
	    (BLURB (NAME) CR |The| |current| |goal| |,| (!PPR NAME (QUOTE |,|))
		   |has| |been| |proved| |,| |and| |has| |no|
		   |dependents| |.| CR)))
	 ; Now return a new state
	 (COND
	   ((SETQ GOAL-AND-INFO (FIND-UNPROVED-GOAL-AND-INFO OTHER-GOALS))
	    (SETQ NEW-NAME (ACCESS GOAL GOAL-NAME (CAR GOAL-AND-INFO)))
	    (BLURB (NEW-NAME) |Now| |proving| (!PPR NEW-NAME (QUOTE |.|))CR)
	    (SWAP-GOALS STATE NEW-NAME GOAL-AND-INFO))
	   (T
	    (BLURB NIL CR |*!*!*!*!*!*!* |
		   |All| |goals| |have| |been|
		   |proved| EXCL |*!*!*!*!*!*!*| CR)
	    (COND
	      ((and (NOT IN-PROVE-LEMMA-FLG)
		    (NOT IN-DEFN-FLG))
	       (BLURB NIL
		      |You| |may| |wish| |to| EXIT |--| |type| |(HELP EXIT)| |for|
		      |details| |.| CR)))
	    STATE))))
      (T STATE))))

(DEFUN CHK-ACCEPTABLE-HINTS-ABB (HINTS ABBREVIATIONS)
  ;; Same as code for CHK-ACCEPTABLE-HINTS except that I translate with
  ;; respect to abbreviations.
  (COND ((OR (AND (ATOM HINTS)
		  HINTS)
	     (CDR (OUR-LAST HINTS)))
	 (ER SOFT (HINTS) |The| HINTS |argument| |must|
	     |be| |a| |proper| |list| |but| (!PPR HINTS NIL) |is| |not| |.|)))
	 
  (ITERATE FOR X IN HINTS
           WITH HINT-NAMES
           DO
           (COND
            ((OR (ATOM X)
		 (CDR (OUR-LAST X)))
             (ER SOFT (X) |each| |element| |of| |the| HINTS |argument| 
                 |must| |be| |a| |non-NIL| |proper| |list| |but|
		 (!PPR X NIL) |is| |not| |.|))
            ((MEMBER-EQ (CAR X) HINT-NAMES)
             (ER SOFT ((Y (CAR X))) |There| |are| |two| (!PPR Y NIL) |hints| EXCL))
            (T (SETQ HINT-NAMES (CONS (CAR X) HINT-NAMES)))))
  (COND
   ((AND (EQUAL (ASSOC-EQ (QUOTE ENABLE-THEORY) HINTS) (QUOTE (ENABLE-THEORY T)))
         (EQUAL (ASSOC-EQ (QUOTE DISABLE-THEORY) HINTS) (QUOTE (DISABLE-THEORY T))))
    (ER SOFT NIL |It| |does| |not| |make| |sense| |both| |to| |enable| |everything|
        |and| |to| |disable| |everything| |,| |as| |you| |have| |done| |with| |the|
        |hints| |(ENABLE-THEORY T)| |and| |(DISABLE-THEORY T)| EXCL)))
  (LIST
   (ITERATE
    FOR X IN HINTS WITH TERM
    COLLECT
      (CASE
       (CAR X)
       (USE
        (CONS (QUOTE USE)
              (ITERATE FOR PAIR IN (CDR X)
                       WITH ALIST
                       COLLECT
                       (COND
                        ((AND (CONSP PAIR)
                              (SYMBOLP (CAR PAIR))
                              (FORMULA-OF (CAR PAIR))
                              (NULL (CDR (OUR-LAST PAIR)))
                              (PROGN
                                (SETQ ALIST
                                      (ITERATE FOR X IN (CDR PAIR)
                                               COLLECT
                                               (COND ((AND (CONSP X)
                                                           (CONSP (CDR X))
                                                           (NULL (CDDR X)))
                                                      (LIST
                                                       (TRANSLATE (CAR X))
                                                       (TRANSLATE-AND-CHK-ABB
                                                        (CADR X)
                                                        ABBREVIATIONS)))
                                                     (T X))))
                               T)
                              (ITERATE FOR X IN ALIST
                                       ALWAYS
                                       (AND (CONSP X)
                                            (CONSP (CDR X))
                                            (NULL (CDDR X))
                                            (VARIABLEP (CAR X)))))
                         (CONS (CAR PAIR) ALIST))
                        (T (ER SOFT
                               ((H (QUOTE
                                    (USE (|event1| (|v1| |t1|) |...|
                                          (|vn| |tn|))
                                         |...|
                                         (|eventk| (|vk| |tk|) |...|
                                          (|vm| |tm|)))))
                                PAIR)
                               |the| USE |hint| |must| |have| |the| |form|
                               (!PPR H NIL) |where| |each| |eventi| |is| |the|
                               |name| |of| |an| ADD-AXIOM |,| CONSTRAIN |,|
			       DEFN |,| FUNCTIONALLY-INSTANTIATE |,| |or|
			       PROVE-LEMMA |event| |,| |each| |vi| |is| |a|
			       |variable| |name| |,| |and| |each| |ti| |is|
			       |a| |term| |.| |the| |entry|
                               (!PPR PAIR NIL) |is| |thus| |unacceptable|
			       |.|))))))
       (EXPAND
        (CONS (QUOTE EXPAND)
              (ITERATE FOR X IN (CDR X)
                       COLLECT
                       (PROGN (SETQ X (TRANSLATE-AND-CHK-ABB X ABBREVIATIONS))
                              (COND ((AND (NVARIABLEP X)
                                          (NOT (FQUOTEP X))
                                          (GET (FFN-SYMB X) (QUOTE SDEFN)))
                                     X)
                                    (T (ER SOFT (X) |every| |element| |of| |an|
                                           EXPAND |hint| |must| |be| |an|
                                           |application| |of| |a| |defined|
                                           |function| |to| |some|
                                           |arguments| |and|
                                           (!TERM X NIL) |is| |not| |.|)))))))
       ((DISABLE ENABLE)
        (ITERATE FOR X IN (CDR X) DO (CHK-DISABLEABLE X))
        X)
       ((DISABLE-THEORY ENABLE-THEORY)
        (COND ((NULL (CDR X))
               (ER SOFT ((Y (CAR X))) |The| (!PPR Y NIL) |hint| |must| |take|
                   |at| |least| |one| |argument| |.|))
              ((MEMBER-EQ T (CDR X))
               (OR (NULL (CDDR X))
                   (ER SOFT ((Y (CAR X))) |The| (!PPR Y NIL) |hint| |must| |take|
                       |either| |the| |single| |argument| |T| |or| |else| |a|
                       |list| |of| |names| |.|)))
              (T
               (ITERATE FOR NAME IN (CDR X) DO
                        (OR (THEORYP NAME)
                            (ER SOFT (NAME) (!PPR NAME NIL) |is| |neither| |GROUND-ZERO|
                                |nor| |the| |name| |of|
                                |a| DEFTHEORY |event| |.|)))))
        X)
       (INDUCT
        (COND ((AND (CONSP (CDR X))
                    (NULL (CDDR X))
                    (PROGN (SETQ TERM (TRANSLATE (CADR X))) T)
                    (NVARIABLEP TERM)
                    (NOT (FQUOTEP TERM))
                    (GET (FFN-SYMB TERM) (QUOTE JUSTIFICATIONS))
                    (GET (FFN-SYMB TERM) (QUOTE INDUCTION-MACHINE))
                    (GET (FFN-SYMB TERM) (QUOTE SDEFN))
                    (ITERATE FOR X IN (FARGS TERM) ALWAYS (VARIABLEP X))
                    (NO-DUPLICATESP (FARGS TERM)))
               (LIST (QUOTE INDUCT) TERM))
              (T (ER SOFT
                     ((G (QUOTE (INDUCT NIL)))
                      (H (QUOTE (INDUCT (|fn| |v1| |...| |vn|))))
                      X)
                     |The| INDUCT |hint| |must| |have| |either| |the| |form|
                     (!PPR G NIL) |or| |the| |form| (!PPR H NIL) |where|
                     |fn| |is| |a| |recursively| |defined| |function| |and|
                     |the| |vi| |are| |distinct| |variables| |.| |Thus| |,|
                     (!PPR X NIL) |is| |an| |inappropriate| INDUCT |hint|
                     |.|))))
       (OTHERWISE
        (COND ((ASSOC-EQ (CAR X) HINT-VARIABLE-ALIST)
               (COND ((CADDR (ASSOC-EQ (CAR X) HINT-VARIABLE-ALIST))
                      (CONS (CAR X)
                            (ITERATE FOR Y IN (CDR X)
                                     COLLECT (TRANSLATE-AND-CHK-ABB Y ABBREVIATIONS))))
                     (T X)))
              (T
               (ER SOFT (X
                         (Y (NCONC (ITERATE FOR X IN HINT-VARIABLE-ALIST
					    COLLECT (CAR X))
                                   '(ENABLE-THEORY DISABLE-THEORY USE INDUCT))))
                   (!PPR X NIL) |is| |an| |unrecognized|
                   |hint| |.| |the| |recognized| |hints|
                   |are| (!LIST Y) |.|))))))))

(DEFUN CHK-ACCEPTABLE-INSTR (INSTR)
; This function returns T or NIL according to whether or
; not the given INSTR is a potentially legal instruction.  If it
; returns NIL then it also side-effects the reason to the screen.
; NOTE:  This function assumes that INSTR is not a positive integer.  It
; is perfectly happy with one-element lists.
  (LET ((INSTR-INFO (GET-INSTR-INFO (instr-name INSTR))))
       (COND
	((INTEGERP INSTR)		;so INSTR is a negative number or 0
	 (BLURB NIL CR |integer| |instructions| |must| |be| |positive| |.| CR))
	((NOT (OR (SYMBOLP INSTR)
		  (CONSP INSTR)))
	 (BLURB NIL |**SORRY**| |--| |Only| |positive| |integers,| |symbols,| |and|
		|proper| |lists| |can| |be| |accepted| |as| |instructions|
		|.| CR))
	((NULL INSTR-INFO)
	 (BLURB ((INSTR-NAME (INSTR-NAME INSTR))) (!PPR INSTR-NAME NIL)
		|is| |not| |a| |valid| |instruction| |.| CR))
	((SYMBOLP INSTR)
	 (COND
	  ((AND (CADR INSTR-INFO)
					;empty list means any number of arguments is OK
		(NOT (MEMBER-EQUAL 0 (CADR INSTR-INFO))))
	   (BLURB (INSTR) CR |the| |instruction| (!PPR INSTR NIL) |is| |not|
		  |allowed| |zero| |arguments| |.| CR))
	  (T T)))
					;hence (CONSP INSTR)
	((CDR (OUR-LAST INSTR))
	 (BLURB NIL CR |no| |improperly| |terminated| |lists| |as|
		|instructions,| |please| |.| CR))
	((AND (CADR INSTR-INFO)		;arity check required for INSTR
	      (NOT (MEMBER-EQUAL (LENGTH (CDR INSTR)) (CADR INSTR-INFO))))
	 (BLURB ((INSTR-NAME (INSTR-NAME INSTR)) (INSTR-ARITIES (CADR INSTR-INFO)))
		CR (!PPR INSTR-NAME NIL) |takes| 
		|only| |a| |number| |of| |arguments| |from| |the| |list|
		(!PPR INSTR-ARITIES (QUOTE |.|)) CR))
	(T T))))

(defun chk-acceptable-interactive-defn-hint (hint)
  (cond
   ((and hint
	 (not (and (match hint (cons 'instructions (cons & &))))))
    (er soft nil |The| |optional| |fifth| |argument| |to| DEFN |must| |be| |a|
	|list| |starting| |with| |the| |atom| INSTRUCTIONS |and| |followed| |by|
	|one| |or| |more| |instructions| |.|)))
  (list hint))

(DEFUN CHK-ACCEPTABLE-INTERACTIVE-HINTS (HINTS FINISH-FLG)

; causes an error if there is an obvious flaw

; First, we must fit into one of the possible cases

  (OR
    ;Start from the beginning
    (AND (NULL (ASSOC-EQ (QUOTE PREVIOUS-EVENT) HINTS))
	 (OR
	   FINISH-FLG		;Finish with something proved
	   (ASSOC-EQ (QUOTE START-GOAL) HINTS)	;A starting point must be specified
	   (ER SOFT NIL
	       |a| |starting| |goal| |must| |be| |specified| |if| |no| |theorem|
	       |is| |proposed| |for| |the| |interactive| PROVE-LEMMA |event| |.|)))
    ;Start from a previous event
    (OR (NULL (ASSOC-EQ (QUOTE START-GOAL) HINTS))	;A starting goal may not be specified
	(ER SOFT NIL
	    |a| |starting| |goal| |may| |not| |be| |specified| |when|
	    |a| |previous| |event ||is| |given| |,| |since| |that| |previous|
	    |state| |stack| |will| |supply| |the| |starting| |point| |for|
	    |the| |current| |sequence| |of| |instructions| |.|)))

  (LIST
    (ITERATE FOR X IN HINTS COLLECT
      (COND
	((ATOM X)
	 (ER SOFT (X) |each| |element| |of| |the| HINTS |argument| |to|
	     PROVE-LEMMA |must| |be| |a| |pair| |but| (!PPR X NIL) |is|
	     |not| |.|))
	(T
	 (CASE
	   (CAR X)
	   (INSTRUCTIONS
	    (OR (CDR X)
		(ER SOFT
		    ((H (QUOTE
			  |(INSTRUCTIONS I1 I2 ... IK)|)))
		    |the| INSTRUCTIONS |hint| |must| |have| |the| |form|
		    (!PPR H NIL) |where| |K>0.|))
	    X)
	   (PREVIOUS-EVENT
	    (COND
	     ((AND (CONSP (CDR X))
		   (NULL (CDDR X))
		   (MEMBER-EQ (CADR X) CHRONOLOGY)
		   (MATCH (GET (CADR X) (QUOTE EVENT))
			  (PROVE-LEMMA & & (QUOTE '*1*TRUE) &))
		   (ASSOC-EQ 'INSTRUCTIONS (NTH 4 (GET (CADR X) (QUOTE EVENT)))))
	      (COND 
	       ((NOT (ASSOC-EQ (CADR X) STATE-STACK-LIST))
		(ER HARD ((PREVIOUS-EVENT (CADR X))) (!PPR PREVIOUS-EVENT NIL) |is| |missing|
		    |from| STATE-STACK-LIST EXCL))
	       ((NULL (CDR (ASSOC-EQ (CADR X) STATE-STACK-LIST)))
		(ER SOFT ((PREVIOUS-EVENT (CADR X))) |Although| |the| |event| (!PPR PREVIOUS-EVENT NIL)
		    |is| |an| |interactive| |PROVE-LEMMA| |event| |with| |a| |partial| |proof| |,|
		    |,| |its| |partial| |proof|
		    |is| |not| |around| |.| |That| |is| |because| |this| |session| |began| |by|
		    |NOTE-LIBing| |a| |library| |which| |someone| |created| |with| |the|
		    |request| |not| |to| |save| |this| |partial| |proof| |.|))))
	     (T
	      (ER SOFT ((H (QUOTE (PREVIOUS-EVENT EV))))
		  |the| PREVIOUS-EVENT |hint| |must| |have| |the| |form|
		  (!PPR H NIL) |where| EV |is| |the| |name| |of| |an|
		  |existing| PROVE-LEMMA |event| |which| |has| |a| |partially-completed|
		  |interactive| |proof| |.|)))
	    X)
	   (START-GOAL
	    (COND
	      ((AND (CONSP (CDR X))
		    (NULL (CDDR X)))
	       (LIST (QUOTE START-GOAL) (TRANSLATE-AND-CHK (CADR X))))
	      (T
	       (ER SOFT ((H (QUOTE (START-GOAL TERM))))
		   |the| START-GOAL |hint| |must| |be| |a| |list|
		   |of| |the| |form| (!PPR H (QUOTE |.|))))))
	   (OTHERWISE
	    (ER SOFT (X
		       (Y (QUOTE (PREVIOUS-EVENT START-GOAL INSTRUCTIONS))))
		(!PPR X NIL) |is| |an| |unacceptable| |hint| |to| PROVE-LEMMA
		|when| |one| |of| |the| |hints| |is| INSTRUCTIONS |(thus| |putting|
		|the| |prover| |in| |interactive| |mode)| |.| |The| |correct|
		|hints| |must| |each| |be| |one| |of| (!LIST Y) |.|))))))))

(DEFUN CLEANUP-CLAUSE (CL ACC)
; simplifies the disjunction of the union of CL and ACC by successively
; moving the members of CL to ACC (and reversing ACC at the end)
  (COND
    ((NULL CL)
     (REVERSE ACC))
    ((OR (MEMBER-EQUAL (CAR CL) ACC)
	 (EQUAL (CAR CL) FALSE))
     (CLEANUP-CLAUSE (CDR CL) ACC))
    ((EQUAL (CAR CL) TRUE)
     (LIST TRUE))
    ((ITERATE FOR TERM IN ACC
	   WITH X
	   THEREIS
	     (OR (AND (MATCH TERM (NOT X))
		      (EQUAL X (CAR CL)))
		 (AND (MATCH (CAR CL) (NOT X))
		      (EQUAL X TERM))))
     (LIST TRUE))
    (T
     (CLEANUP-CLAUSE (CDR CL) (CONS (CAR CL) ACC)))))

(DEFUN CLOSE-ONCE (NEWNAMES GOAL-NAME OTHER-GOALS)
; This adds NEWNAMES to the DEPENDS-ON field of each goal in OTHER-GOALS
; which already depends on GOAL-NAME.  The value returned is a list
; of 4-tuples just like the input OTHER-GOALS.
  (ITERATE FOR GL-INFO IN OTHER-GOALS
	WITH TEMP
	COLLECT
	  (COND
	    ((MEMBER-EQUAL GOAL-NAME (SETQ TEMP (ACCESS GOAL DEPENDS-ON (CAR GL-INFO))))
	     (CONS
	       (MAKE-NEW-RECORD
		 (QUOTE GOAL)
		 (LIST (CONS (QUOTE DEPENDS-ON)
			     (UNION-EQUAL NEWNAMES TEMP)))
		 (CAR GL-INFO))
	       (CDR GL-INFO)))
	    (T GL-INFO))))

(DEFUN CNF (TERM)
; returns a (possibly empty) list of non-empty lists, where NIL
; represents the empty conjunction (i.e., TRUE)
  (ITERATE FOR CL IN (MY-CNF-DNF TERM (QUOTE C))
	WITH NEWCL
	WHEN (NOT (EQUAL (SETQ NEWCL (CLEANUP-CLAUSE CL NIL))
			 (LIST TRUE)))
	  COLLECT
	    (COND
	      ((NULL NEWCL)
	       (LIST FALSE))
	      (T NEWCL))))

(DEFUN COMPARE-LOCAL-AND-GLOBAL-DISABLED-RULES
  (&AUX (INIT-GLOBAL-DISABLED-RULES (INITIAL-FREWRITE-DISABLED-RULES))
	(INIT-LOCAL-DISABLED-RULES (ACCESS STATE FREWRITE-DISABLED-RULES (CAR STATE-STACK)))
	EXTRA-LOCAL-DISABLED MISSING-LOCAL-DISABLED)
  ;; We're already on a fresh line, and this prints a message indicating
  ;; that there's a discrepancy between the local proof-checker enabled-disabled
  ;; state and the global theorem-prover one, if there is such a discrepancy.
  (WHEN (SETQ EXTRA-LOCAL-DISABLED
	      (SET-DIFF-EQ INIT-LOCAL-DISABLED-RULES INIT-GLOBAL-DISABLED-RULES))
	(ER WARNING (EXTRA-LOCAL-DISABLED) |The| |following|
	    (PLURAL? EXTRA-LOCAL-DISABLED |are| |is|) |disabled| |in|
	    |the| |current| |proof-checker| |state|
	    |(i.e.| |for| |commands| |X| |and| |S)| |but|
	    (PLURAL? EXTRA-LOCAL-DISABLED |are| |is|) |enabled| |in| |the|
	    |global| |theorem-prover| |state| |(i.e.|
	    |for| |commands| PROVE |and| |BASH):|
	    (!PPR-LIST EXTRA-LOCAL-DISABLED (QUOTE |.|))))
  (WHEN (SETQ MISSING-LOCAL-DISABLED
	      (SET-DIFF-EQ INIT-GLOBAL-DISABLED-RULES INIT-LOCAL-DISABLED-RULES))
	(ER WARNING (MISSING-LOCAL-DISABLED) |The| |following|
	    (PLURAL? MISSING-LOCAL-DISABLED |are| |is|) |disabled| |in|
	    |the| |global| |theorem-prover| |state| |(i.e.| |for| |commands|
	    PROVE |and| |BASH)| |but|
	    (PLURAL? MISSING-LOCAL-DISABLED |are| |is|) |enabled| |in| |the|
	    |the| |current| |proof-checker| |state|
	    |(i.e.| |for| |commands| |X| |and| |S):|
	    (!PPR-LIST MISSING-LOCAL-DISABLED (QUOTE |.|)))))

(DEFUN COMPILE-UNCOMPILED-MACRO-FUNCTIONS (NAMES)
  (ITERATE FOR NAME IN NAMES
	   WITH FLG
	   WHEN (NOT (COMPILED-FUNCTION-P
		      (SYMBOL-FUNCTION
		       (SETQ NAME (PACK (LIST *PC-MACRO*-PREFIX NAME))))))
	   DO (PROGN (SETQ FLG T) (PRINT NAME) (COMPILE NAME))
	   FINALLY
	   (COND ((NULL FLG)
		  (BLURB NIL CR |Nothing| |to| |compile| |.| CR)))))

(DEFUN CONJUNCTS-OF (TERM)
; See also IMMEDIATE-CONJUNCTS-OF, which doesn't flatten
; the AND structure
  (LET (X Y)
    (IF (MATCH TERM (AND X Y))
	(UNION-EQUAL (CONJUNCTS-OF X) (CONJUNCTS-OF Y))
	(LIST TERM))))

(DEFUN DEL-ASSQ (X L)
; uses EQ for the equality test
  (COND ((ATOM L) NIL)
	((EQ X (CAAR L)) (CDR L))
	(T (CONS (CAR L) (DEL-ASSQ X (CDR L))))))

(DEFUN DEMORGANIZE (TERM)
; The following returns a new term which is logically equivalent to the
; given term but pushes each NOT down through each OR and AND.
  (LET (X Y)
    (COND
      ((MATCH TERM (NOT (OR X Y)))
       (FCONS-TERM* (QUOTE AND)
		    (DEMORGANIZE (DUMB-NEGATE-LIT X))
		    (DEMORGANIZE (DUMB-NEGATE-LIT Y))))
      ((MATCH TERM (NOT (AND X Y)))
       (FCONS-TERM* (QUOTE OR)
		    (DEMORGANIZE (DUMB-NEGATE-LIT X))
		    (DEMORGANIZE (DUMB-NEGATE-LIT Y))))
      (T TERM))))

(DEFUN DEPOSIT-TERM (TERM ADDR SUBTERM)

  ;; Puts SUBTERM at address ADDR in TERM.  Assumes that error checking is
  ;; not necessary, i.e. that ADDR is given correctly relative to TERM,
  ;; sort of (i.e. gives a hard error if not -- someday I may eliminate this
  ;; error checking)

  (COND ((NULL ADDR) SUBTERM)
	((AND (NVARIABLEP TERM)
	      (INTEGERP (CAR ADDR))
	      (NOT (ZEROP (CAR ADDR)))
	      (< (CAR ADDR) (1+ (ARITY (FN-SYMB TERM)))))
	 (CONS-TERM (FN-SYMB TERM)
		    (DEPOSIT-TERM-LST (SARGS TERM) (CAR ADDR) (CDR ADDR) SUBTERM)))
	(T (ER HARD (ADDR (U-W-TERM (UNTRANSLATE-WEAK TERM)))
	       (!PPR ADDR NIL) |is| |an|  |illegal| |address|
	       |into| (!PPR U-W-TERM (QUOTE |.|))))))

(DEFUN DEPOSIT-TERM-LST (LST N ADDR SUBTERM)

; This simply puts (DEPOSIT-TERM term ADDR SUBTERM) in place of the Nth
; element, term, of LST, but avoids consing up the unchanged tail.

  (COND ((INT= 1 N)
	 (CONS (DEPOSIT-TERM (CAR LST) ADDR SUBTERM) (CDR LST)))
	(T (CONS (CAR LST) (DEPOSIT-TERM-LST (CDR LST) (1- N) ADDR SUBTERM)))))

(DEFUN DEPOSIT-TERM-R (TERM ADDR-R SUBTERM)
  (DEPOSIT-TERM TERM (REVERSE ADDR-R) SUBTERM))

(DEFUN DOMAIN (ALIST)
  (ITERATE FOR PAIR IN ALIST COLLECT (CAR PAIR)))

(DEFUN DUMB-DISJOIN (L)
  ;; slight modification of earlier version, adapted from Boyer and Moore
  (COND ((NULL L) FALSE)
	((NULL (CDR L)) (CAR L))
	(T (XXXJOIN 'OR L))))

(DEFUN EVAL-ARG (ARG BINDINGS)
  (LET (X)
    (IF (MATCH ARG (! X))
	(EVAL-LOCAL X BINDINGS)
	(IF (AND (CONSP ARG)
		 (EQ (CAR ARG) (QUOTE !)))
	    (ER SOFT (ARG (EX (QUOTE !))) |The| |argument| (!PPR ARG NIL) |is| |a|
		|list| |which| |begins| |with| |the| |symbol| (!PPR EX NIL)
		|and| |yet| |does| |not| |have| |length| |two| |.|)
	    ARG))))

(DEFUN EVAL-INSTR-ARGS (CDR-INSTR)
  (ITERATE FOR ARG IN CDR-INSTR
	   COLLECT (EVAL-ARG ARG COMMAND-BINDINGS)))

(DEFUN EVAL-LOCAL (FORM BINDINGS)
  (EVAL
    `(LET ,(ITERATE FOR PAIR IN BINDINGS
		 WITH ACC
		 WHEN (NOT (MEMBER-EQ (CAR PAIR) ACC))
		   COLLECT (PROGN (SETQ ACC (CONS (CAR PAIR) ACC))
				  (LIST (CAR PAIR) (KWOTE (CADR PAIR)))))
       ,FORM)))

(DEFUN EXPAND-FNCALL (TERM TYPE-ALIST ID-IFF FREWRITE-RULES FREWRITE-DISABLED-RULES N)
; The following is fixed so that the arguments are simplified too.  Unfortunately,
; this doesn't fix the problem of an outer-level IF giving extra information.
; It assumes that TYPE-ALIST is already fleshed out.
  (LET (FN DEF (TYPE-ALIST1 TYPE-ALIST))
    (COND ((AND (NVARIABLEP TERM)
		(NOT (FQUOTEP TERM))
		(SETQ DEF (GET-DEFN (SETQ FN (FFN-SYMB TERM)))))
	   (SETQ MY-ALL-LEMMAS-USED (ADD-TO-SET FN MY-ALL-LEMMAS-USED))
	   (FREWRITE (CADR DEF)
		     (ITERATE FOR V IN (CAR DEF) AS ARG IN (FARGS TERM)
			   COLLECT (CONS V
					 (FAST-REWRITE ARG TYPE-ALIST1
						       (QUOTE ID)
						       FREWRITE-RULES
						       FREWRITE-DISABLED-RULES N)))
		     TYPE-ALIST1 ID-IFF
		     FREWRITE-RULES FREWRITE-DISABLED-RULES N))
	  (T TERM))))

(DEFUN EXPAND-FNCALL-TAUT (TERM)
  (LET (FN DEF)
    (COND ((AND (NVARIABLEP TERM)
		(NOT (FQUOTEP TERM))
		(SETQ FN (FFN-SYMB TERM))
		(GET FN (QUOTE SDEFN))
		(SETQ DEF
		      (GET-DEFN FN)))
	   (SUB-PAIR-VAR (CAR DEF) (FARGS TERM) (CADR DEF)))
	  (T TERM))))

(DEFUN EXPAND-FNS (TERM FN-LIST EXPAND-LIST)
  
; The following is similar to EXPAND-BOOT-STRAP-NON-REC-FNS, but it lets
; (requires) one to specify a list of functions to be expanded and a list
; of terms whose function calls are to be expanded.  We assume that
; each such function has a definition.  No call is then further expanded, and
; matching with EXPAND-LIST is determined before the arguments are
; rewritten.  Notice that the global MY-ALL-LEMMAS-USED is appropriately
; updated.
  
  (LET (DEF)
    (COND ((VARIABLEP TERM) TERM)
	  ((FQUOTEP TERM) TERM)
	  ((OR (MEMBER-EQ (FFN-SYMB TERM) FN-LIST)
	       (MEMBER-EQUAL TERM EXPAND-LIST))
	   (SETQ MY-ALL-LEMMAS-USED (ADD-TO-SET (FFN-SYMB TERM) MY-ALL-LEMMAS-USED))
	   (SUB-PAIR-VAR (CAR (SETQ DEF (GET-DEFN (FFN-SYMB TERM))))
			 (ITERATE FOR ARG IN (FARGS TERM)
			       COLLECT (EXPAND-FNS ARG FN-LIST EXPAND-LIST))
			 (CADR DEF)))
	  (T
	   (CONS-TERM (FFN-SYMB TERM)
		      (ITERATE FOR ARG IN (FARGS TERM)
			    COLLECT (EXPAND-FNS ARG FN-LIST EXPAND-LIST)))))))

(DEFUN EXTRACT-FREWRITE-RULES (NAMES)
  (ITERATE FOR FN-RULES IN ALL-FREWRITE-RULES
	   COLLECT
	   (CONS (CAR FN-RULES)
		 (ITERATE FOR RULE IN (CDR FN-RULES)
			  WHEN (MEMBER-EQ (CADR RULE) NAMES)
			  COLLECT RULE))))

(DEFUN EXTRACT-PRIMITIVE-RULES (FN-RULES-LIST)
  (ITERATE FOR FN-RULES IN FN-RULES-LIST
	   WITH TEMP
	   WHEN (SETQ TEMP (ITERATE FOR RULE IN (CDR FN-RULES)
				    WHEN (PRIMITIVE-FREWRITE-RULEP RULE)
				    COLLECT RULE))
	   COLLECT (CONS (CAR FN-RULES) TEMP)))

(DEFUN FETCH-TERM-AND-CL (TERM ADDR CL FLAGS)
  
; Returns the subterm of TERM at address ADDR paired
; with a list containing the tests
; governing that occurrence of the subterm plus the literals
; of the input CL, which is actually a list of pairs
; (CONS literal active-flag).  The argument FLAGS is
; a list of flags to be paired, in order, with the governors
; corresponding to the list ADDR; for example, the first flag
; is paired with the first governor picked up by diving
; according to the first address in ADDR, and so on.  (This argument
; is useful when executing the instruction UP, for
; example.)
  
; I've assumed that the address is a list of positive integers.
; If the address is not valid for diving into TERM according to ADDR,
; then NIL is returned.  Notice that ADDR is expected to be in the
; correct order, while CL is in reverse order (like the GOVERNORS) and
; the extension of CL returned in the CDR is also in reverse order.
  
; One could modify a global FETCH-TERM-AND-CL-ERROR in case
; there is an error (as indicated in a comment at the end of the
; code below), but then that global needs to be declared.
  
  (COND ((NULL ADDR)
	 (CONS TERM CL))
	((AND (NVARIABLEP TERM)
	      (< (CAR ADDR) (1+ (ARITY (FN-SYMB TERM)))))
	 (COND ((EQ (FN-SYMB TERM) (QUOTE IF))
		(LET (T1 T2 T3
		      (FLG (IF FLAGS (CAR FLAGS) (QUOTE A)))
		      (NEWFLAGS (IF FLAGS (CDR FLAGS) NIL)))
		  (MATCH TERM (IF T1 T2 T3))
		  (COND ((INT= 1 (CAR ADDR))
			 (FETCH-TERM-AND-CL T1 (CDR ADDR) CL FLAGS))
			((INT= 2 (CAR ADDR))
			 (FETCH-TERM-AND-CL
			   T2
			   (CDR ADDR)
			   (CONS (CONS T1 FLG)
				 CL)
			   NEWFLAGS))
			(T (FETCH-TERM-AND-CL
			     T3
			     (CDR ADDR) 
			     (CONS (CONS (DUMB-NEGATE-LIT T1) FLG) CL)
			     NEWFLAGS)))))
	       (T (FETCH-TERM-AND-CL (NTH (1- (CAR ADDR))
					  (SARGS TERM))
				     (CDR ADDR)
				     CL FLAGS))))
	(T			;(SETQ FETCH-TERM-AND-CL-ERROR (CONS TERM ADDR))
	 NIL)))

(DEFUN FIND-GOAL (NAME ALL-GOALS)
; finds the goal with the given name in ALL-GOALS if there is one;
; otherwise returns NIL
  (ITERATE FOR GL IN ALL-GOALS
	WHEN (EQUAL (ACCESS GOAL GOAL-NAME GL) NAME)
	  DO (RETURN GL)))

(DEFUN FIND-INDUCTION-TERM (FN TERM)
  (COND
    ((VARIABLEP TERM)
     NIL)
    ((FQUOTEP TERM)
     NIL)
    ((AND (EQ (FFN-SYMB TERM) FN)
	  (ITERATE FOR V IN (FARGS TERM) ALWAYS (VARIABLEP V))
	  (NO-DUPLICATESP (FARGS TERM)))
     TERM)
    (T
     (ITERATE FOR X IN (FARGS TERM)
	   WITH TEMP
	   WHEN (SETQ TEMP (FIND-INDUCTION-TERM FN X))
	     DO (RETURN TEMP)))))

(DEFUN FIND-NEXT-BRANCH (TERM)
  
; This function returns a pair (addr . subterm), where  subterm  is one
; which is not an IF-expression and is not a non-FALSE explicit value term,
; i.e. which is a branch and:  is not an explicit value (or is F).
; If no such subterm exists then NIL is returned.  Notice that  addr  comes
; out in the correct order.

  (LET (TEMP)
    (COND
      ((VARIABLEP TERM)
       (CONS NIL TERM))
      ((EQUAL TERM FALSE)
       (CONS NIL TERM))
      ((FQUOTEP TERM)
       NIL)
      ((NEQ (FFN-SYMB TERM) (QUOTE IF))
       (CONS NIL TERM))
      ((SETQ TEMP (FIND-NEXT-BRANCH (FARGN TERM 2)))
       (CONS (CONS 2 (CAR TEMP)) (CDR TEMP)))
      ((SETQ TEMP (FIND-NEXT-BRANCH (FARGN TERM 3)))
       (CONS (CONS 3 (CAR TEMP)) (CDR TEMP)))
      (T NIL))))

(DEFUN FIND-UNPROVED-GOAL-AND-INFO (OTHER-GOALS)
; assuming OTHER-GOALS is a list of 4-tuples (as in the definition of
; the record STATE), returns the first one it finds which does not
; have a TRUE conclusion.
  (ITERATE FOR GL-INFO IN OTHER-GOALS
	WHEN (NOT (EQUAL
		    (ACCESS GOAL CONC (CAR GL-INFO))
		    TRUE))
	  DO (RETURN GL-INFO)))

(DEFUN FIX-INSTR (INSTR)
; This function should be applied in INTERACTIVELY-REWRITE so
; that positive integer "instructions" are only considered in their
; "official" forms  (DIVE N).
  (IF (INTEGERP INSTR)
      (IF (<= INSTR 0)
	  (ER SOFT (INSTR) |Integer| |instructions| |must| |be| |positive| |,|
	      |but| (!PPR INSTR NIL) |is| |not| EXCL)
	  (LIST (QUOTE DIVE) INSTR))
      INSTR))

(defun fix-level-and-length (sexp level length)
  (fix-level-and-length-1 sexp level length length))

(defun fix-level-and-length-1 (sexp level length max-length)
  (cond
   ((atom sexp) sexp)
   ((= length 0)
    '(|...|))
   ((= level 0)
    '|#|)
   (t
    (cons (fix-level-and-length-1 (car sexp) (1- level) max-length max-length)
	  (fix-level-and-length-1 (cdr sexp) level (1- length) max-length)))))

(DEFUN FLATTEN-ANTECEDENTS (TERM)
  (LET (X Y)
    (COND
      ((MATCH TERM (IMPLIES X Y))
       (MAKE-IMPLICATION
	 (ITERATE FOR HYP IN (CONJUNCTS-OF X)
	       APPEND
		 (FLATTEN-ANDS-IN-LIT (DEMORGANIZE HYP)))
	 Y))
      (T TERM))))

(DEFUN FLESH-OUT-ASSUMPTIONS (ASSUMPTIONS)
  (UNION-EQUAL
    (ITERATE FOR X IN ASSUMPTIONS
	  WITH T1 AND T2
	  APPEND
	    (COND
	      ((MATCH X (EQUAL T1 T2))
	       (LIST (FCONS-TERM* (QUOTE EQUAL) T2 T1)))
	      ((MATCH X (NOT (EQUAL T1 T2)))
	       (LIST (FCONS-TERM* (QUOTE NOT)
				  (FCONS-TERM* (QUOTE EQUAL) T2 T1))))
	      ((MATCH X (NLISTP T1))
	       (LIST (FCONS-TERM* (QUOTE NOT)
				  (FCONS-TERM* (QUOTE LISTP) T1))))
	      ((MATCH X (NOT (NLISTP T1)))
	       (LIST (FCONS-TERM* (QUOTE LISTP) T1)))
	      ((MATCH X (NOT (ZEROP T1)))
	       (LIST (FCONS-TERM* (QUOTE NUMBERP) T1)
		     (FCONS-TERM* (QUOTE NOT)
				  (FCONS-TERM* (QUOTE EQUAL) T1 ZERO))
		     (FCONS-TERM* (QUOTE NOT)
				  (FCONS-TERM* (QUOTE EQUAL) ZERO T1))))))
    ASSUMPTIONS))

(DEFUN FLESH-OUT-TYPE-ALIST (TYPE-ALIST)
  (COND
    ((ATOM TYPE-ALIST)		;i.e., CONTRADICTION
     TYPE-ALIST)
    (T
     (UNION-EQUAL
       (ITERATE FOR PAIR IN TYPE-ALIST
	     WITH T1 AND T2
	     APPEND
	       (COND
		 ((MATCH (CAR PAIR) (EQUAL T1 T2))
		  (LIST (CONS (FCONS-TERM* (QUOTE EQUAL) T2 T1)
			      (CDR PAIR))))
		 ((MATCH (CAR PAIR) (NLISTP T1))
		  (COND
		    ((EQUAL (CDR PAIR) TYPE-SET-TRUE)
		     (LIST (CONS T1 (TSLOGNOT TYPE-SET-CONS))))
		    ((EQUAL (CDR PAIR) TYPE-SET-FALSE)
		     (LIST (CONS T1 TYPE-SET-CONS)))))
		 ((AND (MATCH (CAR PAIR) (ZEROP T1))
		       (EQUAL (CDR PAIR) TYPE-SET-FALSE))
		  (LIST (CONS T1 TYPE-SET-NUMBERS)
			(CONS (FCONS-TERM* (QUOTE EQUAL) T1 ZERO)
			      TYPE-SET-FALSE)
			(CONS (FCONS-TERM* (QUOTE EQUAL) ZERO T1)
			      TYPE-SET-FALSE)))))
       TYPE-ALIST))))

(defun free-vars-to-bind (new-goal-free-vars other-goals free-variables gen-vars)
  ;; Below, this of the parameters as follows:
  ;; gen-vars = (all-vars-lst (mapcar #'car gen-pairs))
  ;; new-goal-free-vars = (intersection-eq free-variables (all-vars-goal new-goal))
  (intersection-eq
   (gen-closure (cons new-goal-free-vars
		      (iterate for gl-info in other-goals
			       collect
			       (intersection-eq free-variables
						(all-vars-goal (car gl-info)))))
		new-goal-free-vars)
   gen-vars))

(DEFUN FREWRITE-RULES-FOR (FN NEW-LEMMA-NAMES)
  ;; NOTE:  It's harmless if NEW-LEMMA-NAMES contains
  ;; names that aren't names of lemmas.
  (IF (EQ NEW-LEMMA-NAMES (QUOTE *FAILED*))
      (ITERATE FOR LEMMA IN (GET FN (QUOTE LEMMAS))
	       WHEN (AND (SETQ LEMMA (MAKE-EQUALITY LEMMA 'ID))
			 (ALMOST-PRIMITIVE-REWRITE-RULEP LEMMA))
	       COLLECT (MAKE-FREWRITE-RULE LEMMA))
      (ITERATE FOR LEMMA IN (GET FN (QUOTE LEMMAS))
	       WHILE (MEMBER-EQ
		      (ACCESS REWRITE-RULE NAME LEMMA)
		      NEW-LEMMA-NAMES)
	       WHEN (AND (SETQ LEMMA (MAKE-EQUALITY LEMMA 'ID))
			 (ALMOST-PRIMITIVE-REWRITE-RULEP LEMMA))
	       COLLECT (MAKE-FREWRITE-RULE LEMMA))))

(defun gen-closure (varlists vars)
  ;; The idea will be to form the gen-closure by starting with a data
  ;; structure that is a list of varlists, and then apply
  ;; gen-closure to that.  Unlike the version I verified, I'll
  ;; assume that I intersected with the list of free vars before
  ;; I started.
  (iterate with indices and result and flg = t
	   while flg
	   do
	   (iterate for varlist in varlists
		    as i from 1
		    initially
		    ;; the progn below is to overcome an apparent sloop bug
		    (progn (setq flg nil))
		    when (and (not (member i indices))
			      (intersectp-eq varlist vars)
			      (setq indices (cons i indices)))
		    do
		    (setq flg t)
		    (setq result (union-eq varlist result)))
	   finally (return result)))

(DEFUN GET-DEFN (FN)
; returns (formals body ...).
  (OR (SAFECDR (SAFECDR (GET FN (QUOTE EVENT))))
      (SAFECDR (GET FN (QUOTE SDEFN)))))

(DEFUN GET-FREWRITE-DISABLEDP (NAME FREWRITE-DISABLED-RULES)
  
; answers the question, is NAME disabled as an frewrite rule.

  (COND ((MEMBER-EQ NAME FREWRITE-DISABLED-RULES) T)
	(T NIL)))

(DEFUN GET-FREWRITE-RULES (FN FREWRITE-RULES)

; fetches the frewrite rules about the function named FN.

  (CDR (ASSOC-EQ FN FREWRITE-RULES)))

(DEFUN GET-QUICK-OPENER (FN)

; fetches the quick opener for the function named FN.

  (ASSOC-EQ FN FREWRITE-QUICK-OPENERS))

(DEFUN GOAL-IMMEDIATE-DEPENDENTS (GOAL GOALS)
  (LET ((DEPENDENTS (ACCESS GOAL DEPENDS-ON GOAL)))
    (SET-DIFF
     DEPENDENTS
     (ITERATE FOR GL IN GOALS
	      WHEN (MEMBER-EQUAL (ACCESS GOAL GOAL-NAME GL) DEPENDENTS)
	      APPEND (ACCESS GOAL DEPENDS-ON GL)))))

(DEFUN HIGHLIGHT-PPRIND (FMLA LEFTMARGIN RPARCNT PPRFILE
			 CURRENT-ADDR-R CURRENT-TERM ABBREVIATIONS)
  (LET ((FMLA0 (UNTRANSLATE-ABB-WEAK FMLA ABBREVIATIONS))
	(TERM0 (UNTRANSLATE-ABB-WEAK CURRENT-TERM ABBREVIATIONS)))
    (PPRIND (DEPOSIT-TERM-R FMLA0 CURRENT-ADDR-R
			    (LIST (QUOTE ***) TERM0 (QUOTE ***)))
	    LEFTMARGIN RPARCNT PPRFILE)))

(DEFUN HYPS-TYPE-ALIST (HYPS)
  (LET (TYPE-ALIST)
    (FLESH-OUT-TYPE-ALIST	;just to ensure fleshing out
      (MY-TYPE-ALIST-CLAUSE
	(ITERATE FOR HYP IN (FLESH-OUT-ASSUMPTIONS (ASSUMPTIONS HYPS))
				;fleshed out here in order to obtain CONTRADICTION
				;when possible
	      COLLECT (DUMB-NEGATE-LIT HYP))))))

(defun immediate-conjuncts-of (term &optional (n 0))
  ;; Maps (and x y ...) to (x y ...) and maps any other term
  ;; tm to (tm).  Except, n is the maximum number of conjuncts expected,
  ;; and hence if n reaches 0 we return the entire AND.
  (let (x y)
       (if (int= n 1)
	   (list term)
	   (if (match term (and x y))
	       (add-to-set x (immediate-conjuncts-of y (1- n)))
	     (list term)))))

(DEFUN IMMEDIATE-CONJUNCTS-OF-ALLOWING-REPEATS (TERM)
; maps (and x y ...) to (x y ...) and maps any other term
; tm to (tm).
  (LET (X Y)
    (IF (MATCH TERM (AND X Y))
	(CONS X (IMMEDIATE-CONJUNCTS-OF-ALLOWING-REPEATS Y))
	(LIST TERM))))

(DEFUN INITIAL-FREWRITE-DISABLED-RULES NIL
  (ITERATE FOR TRIPLE IN DISABLED-LEMMAS
	   WITH TEMP AND ITERATE-ANS
	   WHEN (NOT (MEMBER-EQ (CAR TRIPLE) TEMP)) 
	   DO (PROGN (SETQ TEMP (CONS (CAR TRIPLE) TEMP))
		     (COND ((AND (CDDR TRIPLE)
				 (OR (MEMBER-EQ (CAR TRIPLE)
						ALL-FREWRITE-RULE-NAMES)
				     (MEMBER-EQ (CAR TRIPLE)
						ALL-DEFINED-FUNCTION-NAMES)))
			    (SETQ ITERATE-ANS (CONS (CAR TRIPLE)
						    ITERATE-ANS)))))
	   FINALLY (RETURN ITERATE-ANS)))

(DEFUN INITIALIZE-FREWRITE NIL
  (IF (EQ CHRONOLOGY OLD-CHRONOLOGY)
      NIL
      (LET ((NEW-CITIZEN-NAMES (NEW-CITIZEN-NAMES CHRONOLOGY OLD-CHRONOLOGY))
	    NEW-ALL-FREWRITE-RULES NEW-FREWRITE-RULES
	    NEW-ALL-FREWRITE-RULE-NAMES NEW-ALL-DEFINED-FUNCTION-NAMES
	    NEW-FREWRITE-QUICK-OPENERS
	    ADDITIONAL-ALL-FREWRITE-RULES)
    
	   ;; initializes the lists of rewrite rules for FREWRITE.
	   ;; Also initializes the QUICK-OPENERS for FREWRITE and
	   ;; a couple of lists of names.

	   (SETQ NEW-ALL-FREWRITE-RULES
		 (COND
		  ((NEQ NEW-CITIZEN-NAMES (QUOTE *FAILED*))
		   (UPDATE-FN-RULES-ALIST
		    ALL-FREWRITE-RULES
		    (SETQ ADDITIONAL-ALL-FREWRITE-RULES
			  (ALL-ALMOST-PRIMITIVE-REWRITE-RULES NEW-CITIZEN-NAMES))))
		  (T
		   (ALL-ALMOST-PRIMITIVE-REWRITE-RULES (QUOTE *FAILED*)))))
	   (SETQ NEW-FREWRITE-RULES
		 (COND
		  ((NEQ NEW-CITIZEN-NAMES (QUOTE *FAILED*))
		   (UPDATE-FN-RULES-ALIST
		    FREWRITE-RULES
		    (EXTRACT-PRIMITIVE-RULES ADDITIONAL-ALL-FREWRITE-RULES)))
		  (T
		   (EXTRACT-PRIMITIVE-RULES NEW-ALL-FREWRITE-RULES))))
	   (SETQ NEW-ALL-FREWRITE-RULE-NAMES
		 (ITERATE FOR FN-RULES IN NEW-ALL-FREWRITE-RULES
			  NCONC
			  (ITERATE FOR RULE IN (CDR FN-RULES)
				   COLLECT (CADR RULE))))
	   (SETQ NEW-ALL-DEFINED-FUNCTION-NAMES
		 (COND
		  ((NEQ NEW-CITIZEN-NAMES (QUOTE *FAILED*))
		   (NCONC (ITERATE FOR X IN NEW-CITIZEN-NAMES
				   WHEN (GET X (QUOTE SDEFN))
				   COLLECT X)
			  ALL-DEFINED-FUNCTION-NAMES))
		  (T
		   (NCONC
		    (ITERATE FOR X IN CHRONOLOGY
			     WHEN (GET X (QUOTE SDEFN))
			     COLLECT X)
		    (ITERATE FOR X IN (GET (QUOTE GROUND-ZERO) (QUOTE SATELLITES))
			     WHEN (GET X (QUOTE SDEFN))
			     COLLECT X)))))
	   (SETQ NEW-FREWRITE-QUICK-OPENERS
		 (COND
		  ((NEQ NEW-CITIZEN-NAMES (QUOTE *FAILED*))
		   (NCONC
		    (ITERATE FOR X IN NEW-CITIZEN-NAMES
			     WITH TEMP
			     WHEN (SETQ TEMP (MAKE-QUICK-OPENER X)) ;checks induction machine
			     COLLECT (CONS X TEMP))
		    FREWRITE-QUICK-OPENERS))
		  (T
		   (NCONC
		    (ITERATE FOR X IN CHRONOLOGY
			     WITH TEMP
			     WHEN (SETQ TEMP (MAKE-QUICK-OPENER X))
			     COLLECT (CONS X TEMP))
		    (ITERATE FOR X IN (GET (QUOTE GROUND-ZERO) (QUOTE SATELLITES))
			     WITH TEMP
			     WHEN (SETQ TEMP (MAKE-QUICK-OPENER X))
			     COLLECT (CONS X TEMP))))))
	   ;; Now we make all the changes roughly at once, to get the effect of atomicity
	   (PSETQ ALL-FREWRITE-RULES NEW-ALL-FREWRITE-RULES
		  FREWRITE-RULES NEW-FREWRITE-RULES
		  ALL-FREWRITE-RULE-NAMES NEW-ALL-FREWRITE-RULE-NAMES
		  ALL-DEFINED-FUNCTION-NAMES NEW-ALL-DEFINED-FUNCTION-NAMES
		  FREWRITE-QUICK-OPENERS NEW-FREWRITE-QUICK-OPENERS
		  OLD-CHRONOLOGY CHRONOLOGY))))

(DEFUN INSTR-AND-TYPE (INPUT)
  ; returns the CONS of an instruction with the TYPE (as in C, H,
  ; or M), assuming such an instruction exists; returns NIL otherwise
  (SETQ INPUT (FIX-INSTR INPUT))	;replaces non-negative integer
			                ;instructions in terms of DIVE or UP
  (COND
    ((CHK-ACCEPTABLE-INSTR INPUT)	;print error message here
     (MAKE-INSTR-AND-TYPE INPUT))))

(DEFUN INSTR-NAME (INSTR)
  ;; patched to take care of e.g. (2 2) as an instruction.  Notice that READ-INSTR
  ;; calls this early on
  (IF (CONSP INSTR)
      (IF (SYMBOLP (CAR INSTR))
	  (CAR INSTR)
	  (ER SOFT (INSTR) |Command| |names| |must| |be| |symbols| |,| |so|
	      (!PPR INSTR NIL) |is| |not| |a| |valid| |instruction| EXCL))
      INSTR))

(DEFUN INTERACTIVE-HINTS-P (HINTS)
  (AND (CONSP HINTS)
       ;; some redundancy here with CHK-ACCEPTABLE-HINTS; oh well
       (ITERATE FOR X IN HINTS
		ALWAYS
		(OR (CONSP X)
		    (ER SOFT (X) |Each| |element| |of| |the|
			HINTS |argument| |to| PROVE-LEMMA |must|
			|be| |a| |pair| |but| (!PPR X NIL) |is|
			|not| |.|)))
       (ASSOC-EQ (QUOTE INSTRUCTIONS) HINTS)))

(DEFUN INTERSECTP-EQ (X Y)
  ;; adapted from Boyer-Moore INTERSECTP
  (ITERATE FOR E IN X THEREIS (MEMBER-EQ E Y)))

(DEFUN LEMMA-VARS-HYPS-REMAINING
  (LEMMA SUBST DOLLAR-SUBST ASSUMPTIONS TYPE-ALIST
	       &OPTIONAL DOLLAR-FLG DO-NOT-COMPLETE-FLG)
; -- Returns an object of the form
;      (LEMMA VARS . HYPS)
;   Here LEMMA-SUBST is a member of a list returned by APPLICABLE-REWRITES,
; i.e. a lemma LEMMA with conclusion of the form (EQUAL t1 t2) consed on to
; a substitution SUBST.  The LEMMA returned is the result of putting dollar
; signs in front of each free variable in the lemma.
; VARS is then the free variables of LEMMA, where we
; have put dollar signs in front of each of them, and HYPS is the set of
; hypotheses of the lemma as instantiated by SUBST and by DOLLAR-SUBST.  In
; fact the dollar signs are removed from this list of HYPS that is 
; returned unless DOLLAR-FLG is set.  For, even though we need to do
; such removal anyhow to check which hypotheses are left to be relieved,
; and this is the form we may ultimately want for HYPS when actually
; rewriting, still I'd like to use this code for the
; SHOW-REWRITES command, which I do when DOLLAR-FLG is set.
;    It's not clear that I'll ever use the DO-NOT-COMPLETE-FLG option.

; NOTE HOWEVER that if VARS is non-NIL and the DO-NOT-COMPLETE-FLG is NIL,
;    and also there are no variables free in the resulting t2 (i.e. after
;    substituting with SUBST and then DOLLAR-SUBST),
;    then we make some attempt to complete DOLLAR-SUBST in order to eliminate
;    all hypotheses.  If this attempt is not totally successful then we
;    pretend that we never tried.
; NOTE:  We assume that TYPE-ALIST corresponds to ASSUMPTIONS -- in fact
;    notice that we actually pass it as a parameter.
  (LET ((NEW-LEMMA-AND-FREE-VARS
	 (SHOW-FREE-VARIABLES-IN-LEMMA LEMMA))
	(DOMAIN (DOMAIN DOLLAR-SUBST)))
       (LET ((FREE-VARS (CDR NEW-LEMMA-AND-FREE-VARS))
	     (LEMMA (CAR NEW-LEMMA-AND-FREE-VARS)))
	    (LET
	     ((HYPS
	       (MAKE-SET
		(SUBLIS-VAR-LST (APPEND DOLLAR-SUBST SUBST)
					; it's ok to append above since the domain of
					; DOLLAR-SUBST is disjoint from the range
					; of SUBST
				(ACCESS REWRITE-RULE HYPS LEMMA)))))
	     (COND
	      ((AND FREE-VARS
		    (NOT DO-NOT-COMPLETE-FLG)
		    (ITERATE FOR ATOM
			     IN (ALL-ATOMS (FARGN (ACCESS REWRITE-RULE CONCL LEMMA)
						  2))
			     THEREIS
			     (AND 
			      (EQUAL (CHAR (STRING ATOM) 0) #\$)
			      (NOT (MEMBER-EQ ATOM DOMAIN)))))
	       (SETQ DO-NOT-COMPLETE-FLG T)))
	     (CONS LEMMA
		   (CONS FREE-VARS
			 (COND
			  ((AND
			    FREE-VARS
			    (NOT DO-NOT-COMPLETE-FLG))
			   (MATCH-FREE-VARS HYPS
					    ASSUMPTIONS
					    DOLLAR-SUBST)
			   (IF (NULL
				(TYPE-SET-NON-DEDUCIBLE
				 (UNDOLLAR-FY-LST (SUBLIS-VAR-LST UNIFY-SUBST HYPS))
				 TYPE-ALIST))
			       NIL
			       (TYPE-SET-NON-DEDUCIBLE
				(IF DOLLAR-FLG HYPS (UNDOLLAR-FY-LST HYPS))
				TYPE-ALIST)))
			  (T (TYPE-SET-NON-DEDUCIBLE
			      (IF DOLLAR-FLG HYPS (UNDOLLAR-FY-LST HYPS))
			      TYPE-ALIST)))))))))

(DEFUN LONG-HELP (INSTR-NAME)
  (EVAL (PACK (LIST INSTR-NAME (QUOTE -) (QUOTE LONG-HELP-MESSAGE)))))

(DEFUN MAKE-BLURB-BODY (SEQ)
  (LET ((BINDINGS
	  (ITERATE FOR X IN SEQ
		WITH VARS = (QUOTE (A B C D E G H I J K L M N))
		WHEN (CONSP X)
		  COLLECT
		(PROG1 (LIST (CAR VARS) (KWOTE X))
		       (SETQ VARS (CDR VARS))))))
    (CONS BINDINGS
	  (ITERATE FOR X IN SEQ
		WITH TEMP = BINDINGS
		COLLECT
		(IF (ATOM X)
		    X
		    (PROG1 `(!PPR ,(CAAR TEMP) NIL)
			   (SETQ TEMP (CDR TEMP))))))))

(DEFUN MAKE-COPY-ALIST (COPY-VARS OLD-VARS)
  (ITERATE FOR V IN COPY-VARS
	   WITH NEW-VAR
	   COLLECT
	   (PROGN (SETQ NEW-VAR (MAKE-NEW-COPY-VAR V OLD-VARS))
		  (SETQ OLD-VARS (CONS NEW-VAR OLD-VARS))
		  (CONS V NEW-VAR))))

(DEFUN MAKE-EQUALITY (LEMMA ID-IFF)
; Returns NIL if one has a meta-lemma or if one can't make an equality
; (because the flag is ID and the conclusion isn't boolean).  Otherwise,
; returns a new equivalent (under the ID-IFF flag) rewrite rule whose
; conclusion is an equality.
  (LET ((CONCL (ACCESS REWRITE-RULE CONCL LEMMA)))
    (COND
      ((META-LEMMAP LEMMA)
       NIL)
      ((EQ (FFN-SYMB CONCL)
	   (QUOTE NOT))
       (MAKE-NEW-RECORD
	 (QUOTE REWRITE-RULE)
	 (LIST (CONS (QUOTE CONCL)
		     (FCONS-TERM* (QUOTE EQUAL)
				  (FARGN CONCL 1)
				  FALSE)))
	 LEMMA))
      ((EQ (FFN-SYMB CONCL) (QUOTE EQUAL))
       LEMMA)
      ((EQ (FFN-SYMB CONCL) (QUOTE IFF))
       (IF (OR (EQ ID-IFF (QUOTE IFF))
	       (LET (TYPE-ALIST) (OUR-BOOLEAN (FARGN CONCL 1))))
	   (MAKE-NEW-RECORD
	     (QUOTE REWRITE-RULE)
	     (LIST (CONS (QUOTE CONCL)
			 (FCONS-TERM (QUOTE EQUAL) (FARGS CONCL))))
	     LEMMA)
	   NIL))
      ((OR (EQ ID-IFF (QUOTE IFF))
	   (LET (TYPE-ALIST) (OUR-BOOLEAN CONCL)))
       (MAKE-NEW-RECORD
	 (QUOTE REWRITE-RULE)
	 (LIST (CONS (QUOTE CONCL)
		     (FCONS-TERM* (QUOTE EQUAL) CONCL TRUE)))
	 LEMMA))
      (T NIL))))

(DEFUN MAKE-FORMULA-FROM-CLAUSE (CL)
; here CL is a list to be viewed as a disjunction
  (COND
    ((NULL CL)
     FALSE)
    ((NULL (CDR CL))
     (CAR CL))
    (T
     (MAKE-IMPLICATION
       (ITERATE FOR TAIL ON CL
	     WHILE (CDR TAIL)
	     COLLECT (DUMB-NEGATE-LIT (CAR TAIL)))
       (CAR (OUR-LAST CL))))))

(DEFUN MAKE-FREWRITE-RULE (LEMMA)
  (LIST (QUOTE FREWRITE-RULE)
	(ACCESS REWRITE-RULE NAME LEMMA)
	(ACCESS REWRITE-RULE HYPS LEMMA)
	(FARGN (ACCESS REWRITE-RULE CONCL LEMMA)
	       1)
	(FARGN (ACCESS REWRITE-RULE CONCL LEMMA)
	       2)))

(DEFUN MAKE-HYP-FLGS (HYP-LST)
  (ITERATE FOR HYP IN HYP-LST
	COLLECT (CONS HYP (QUOTE A))))

(DEFUN MAKE-IMPLICATION (ASSUMPTIONS CONCL)
  (COND
    (ASSUMPTIONS
     (FCONS-TERM* (QUOTE IMPLIES) (DUMB-CONJOIN ASSUMPTIONS) CONCL))
    (T CONCL)))

(DEFUN MAKE-INSTR-AND-TYPE (INSTR)
  (CONS INSTR
	(CAR (GET-INSTR-INFO (INSTR-NAME INSTR)))))

(DEFUN MAKE-NEW-COPY-VAR (V OLD-VARS &AUX (VAR-INDEX (PREPARE-FOR-COPY V)))
  (MAKE-NEW-VAR (CAR VAR-INDEX) (CDR VAR-INDEX) OLD-VARS))

(DEFUN MAKE-NEW-IND-INSTANCE (TERM TESTS-AND-CASES FORMALS VARS)
  (LET ((TESTS (ACCESS TESTS-AND-CASES TESTS TESTS-AND-CASES))
	(CASES (ACCESS TESTS-AND-CASES CASES TESTS-AND-CASES))
	HYP)
    (SETQ TESTS
	  (SUB-PAIR-VAR-LST FORMALS VARS TESTS))
    (SETQ CASES
	  (ITERATE FOR CASE IN CASES
		COLLECT
		  (SUB-PAIR-VAR-LST FORMALS VARS CASE)))
    (SETQ HYP
	  (DUMB-CONJOIN
	    (APPEND TESTS
		    (ITERATE FOR CASE IN CASES
			  COLLECT
			    (SUB-PAIR-VAR VARS CASE TERM)))))
    (FCONS-TERM* (QUOTE IMPLIES) HYP TERM)))

(DEFUN MAKE-NEW-RECORD (RECORD-NAME FIELD-VALUE-PAIRS RECORD)
  (SETQ FIELD-VALUE-PAIRS	;put pairs in order, for space efficiency
	(OUR-STABLE-SORT
	  (ITERATE FOR FIELD-VALUE IN FIELD-VALUE-PAIRS
		COLLECT
		  (CONS (GET-FIELD-NUMBER RECORD-NAME (CAR FIELD-VALUE))
			(CDR FIELD-VALUE)))
	  (FUNCTION (LAMBDA (X Y) (< (CAR X) (CAR Y))))))
  (SETQ RECORD			;take care of CHEAP flag
	(IF (CADDR (ASSOC-EQ RECORD-NAME RECORD-DECLARATIONS))
	    RECORD
	    (CDR RECORD)))
  (SETQ RECORD
	(ITERATE FOR COMPONENT IN RECORD AS I FROM 1
	      COLLECT
		(PROGN
		  (COND
		    ((AND FIELD-VALUE-PAIRS
			  (INT= I (CAAR FIELD-VALUE-PAIRS)))
		     (SETQ COMPONENT (CDAR FIELD-VALUE-PAIRS))
		     (SETQ FIELD-VALUE-PAIRS
			   (CDR FIELD-VALUE-PAIRS))))
		  COMPONENT)))
  (IF (CADDR (ASSOC-EQ RECORD-NAME RECORD-DECLARATIONS))
      RECORD
      (CONS RECORD-NAME RECORD)))

(DEFUN MAKE-NEW-VAR (V INDEX OLD &AUX (NEW-V (PACK (LIST V INDEX))))
  (IF (MEMBER-EQ NEW-V OLD)
      (MAKE-NEW-VAR V (1+ INDEX) OLD)
      NEW-V))

(DEFUN MAKE-QUICK-OPENER (FN)

;  A quick-opener for fn is NIL if the fn is nonrecursive and is otherwise
;  a list in 1 to 1 correspondence with the formals of fn.  To each formal v
;  we assign a token x.  If v is, by itself, a measured subset whose COUNT
;  decreases in each call and every call destructs v with an accessor for
;  the same shell constructor const, the token is the shell constructor.  
;  Otherwise the token is NIL.

;  The idea behind a quick-opener is that if the actual in any non-NIL
;  slot has as its fn symb the token for that slot then the call can be opened.
;  (But we don't unless the actual has the wrong type or is a bottom object.)
;  Furthermore, if the actual is of the wrong type, it can be opened too.

  (LET (TEMP)
    (COND ((GET FN (QUOTE INDUCTION-MACHINE))
	   (ITERATE FOR V IN (CADR (GET FN (QUOTE SDEFN))) AS I FROM 0
		 COLLECT
		   (COND ((AND (ITERATE FOR X IN (GET FN (QUOTE JUSTIFICATIONS)) THEREIS
				     (AND (EQUAL (ACCESS JUSTIFICATION SUBSET X)
						 (LIST V))
					  (EQUAL (ACCESS JUSTIFICATION
							 MEASURE-TERM X)
						 (FCONS-TERM* (QUOTE COUNT) V))))
			       (SETQ TEMP
				     (QUICK-OPENER-TYPE
				       V
				       (ITERATE FOR X IN (GET FN (QUOTE INDUCTION-MACHINE))
					     WHEN (ACCESS TESTS-AND-CASES CASES X)
					     DO (RETURN
						  (NTH
						    I
						    (CAR
						      (ACCESS
							TESTS-AND-CASES
							CASES X)))))))
			       (ITERATE FOR X IN (GET FN (QUOTE INDUCTION-MACHINE))
				     ALWAYS (ITERATE FOR ARGS IN
						  (ACCESS TESTS-AND-CASES CASES X)
						  ALWAYS
						  (EQ TEMP
						      (QUICK-OPENER-TYPE
							V
							(NTH I ARGS))))))
			  TEMP)
			 (T NIL))))
	  (T NIL))))

(DEFUN MAKE-SET (L)
;; non-destructive version, used everywhere except in places where it's
;; syntactically obvious that we are only destroying a newly consed-up list.
;; This version uses EQUAL.
  (REMOVE-DUPLICATES L :TEST #'EQUAL))

(DEFUN MATCH-FREE-VARS (TERMS ASSUMPTIONS SUBSTITUTION)
; Here ASSUMPTIONS is a list of hypotheses, TERMS is a list of terms which may
; include some "free variables" (i.e. variables starting with the dollar-sign "$"),
; and our goal is to find a substitution for the "dollar" variables which
; extends the given SUBSTITUTION, such that each of the resulting
; terms is a member of the list ASSUMPTIONS.  We return T if one is found, and
; NIL if not.
  (SETQ UNIFY-SUBST SUBSTITUTION)
  (MATCH-FREE-VARS1 TERMS
		    (FLESH-OUT-ASSUMPTIONS ASSUMPTIONS)))

(DEFUN MATCH-FREE-VARS1 (TERMS ASSUMPTIONS)
; Notice that this deals with the global UNIFY-SUBST, which may already
; be partially formed when this is called.  Notice also that if NIL is
; returned then UNIFY-SUBST is not changed.
; We only call this when TYPE-ALIST corresponds to ASSUMPTIONS.
; We assume that UNIFY-SUBST begins as a substitution only on dollar
; variables, and ends that way.  Thus, if the (NOT (TSLOGSUBSETP ....))
; test succeeds, and hence it is a theorem that the term in question
; is not F if one assumes ASSUMPTIONS, then this property continues to
; hold as UNIFY-SUBST is extended, since no dollar variables occur in
; ASSUMPTIONS.
  (COND
    ((NULL TERMS)
     T)
    ; here's something new for eliminating trivial equalities and such
    ; **** At one point I was a bit worried that it may be noticeably slower than what
    ; I formerly had, which was nothing here.  I thought that maybe I should just check
    ; a match of (EQUAL X Y) where X = Y.  But I'm not aware of any problems along
    ; these lines so I'll just leave it alone.
    ((NOT (TSLOGSUBSETP TYPE-SET-FALSE
		      (MY-TYPE-SET (MY-SUBLIS-VAR UNIFY-SUBST (CAR TERMS)))))
     (MATCH-FREE-VARS1 (CDR TERMS) ASSUMPTIONS))
    (T
     (ITERATE FOR X IN ASSUMPTIONS
	   WITH TEMP-SUBST = UNIFY-SUBST
	   THEREIS
	   (OR
	    (AND (ONE-WAY-UNIFY11-$ (CAR TERMS) X)
		 (MATCH-FREE-VARS1 (CDR TERMS) ASSUMPTIONS))
	    (PROGN (SETQ UNIFY-SUBST TEMP-SUBST)
		   NIL))))))

(DEFUN MY-CNF-DNF (TERM FLG)
  
;   Taken from the definition of CNF-DNF in the Boyer-Moore code, but
;   augmented to handle IF and IFF.  That is, the only differences are in the
;   binding of R in the top LET, the recursive calls to MY-CNF-DNF
;   instead of to CNF-DNF, and the case (MATCH TERM (IF P Q R)).
  
;   If FLG is (QUOTE C), returns a list of lists, say:

;   ((p11 p12 ...) (p21 p22  ...) ...  (pn1 pn2 ...))

;   such that TERM is not equal to F iff

;   (AND (OR p11 p12 ...) (OR p21 p22 ...) ... (OR pn1 pn2 ...))

;   is not equal to F.  The latter term is the "conjunctive normal form" of
;   TERM.  If FLG is (QUOTE D) computes the disjunctive normal form.

  (LET (P Q R NF-Q)
    (COND ((OR (AND (EQ FLG (QUOTE C))
                    (MATCH TERM (AND P Q)))
               (AND (EQ FLG (QUOTE D))
                    (MATCH TERM (OR P Q))))
           (APPEND (MY-CNF-DNF P FLG)
                   (MY-CNF-DNF Q FLG)))
          ((OR (AND (EQ FLG (QUOTE C))
                    (MATCH TERM (OR P Q)))
               (AND (EQ FLG (QUOTE D))
                    (MATCH TERM (AND P Q))))
           (SETQ NF-Q (MY-CNF-DNF Q FLG))
           (ITERATE FOR L1 IN (MY-CNF-DNF P FLG)
                    WITH ITERATE-ANS
                    DO (SETQ ITERATE-ANS
                             (UNION-EQUAL (ITERATE FOR L2 IN NF-Q
                                                   COLLECT (UNION-EQUAL L1 L2))
                                          ITERATE-ANS))
                    FINALLY (RETURN ITERATE-ANS)))
          ((MATCH TERM (NOT P))
           (ITERATE FOR L1 IN (MY-CNF-DNF P (CASE FLG
                                               (D (QUOTE C))
                                               (OTHERWISE (QUOTE D))))
                    COLLECT (ITERATE FOR TERM IN L1
                                     COLLECT (DUMB-NEGATE-LIT TERM))))
          ((MATCH TERM (IF P Q R))
           (MY-CNF-DNF (FCONS-TERM* (QUOTE AND)
                                    (FCONS-TERM* (QUOTE OR) (DUMB-NEGATE-LIT P) Q)
                                    (FCONS-TERM* (QUOTE OR) P R))
                       FLG))
          ((MATCH TERM (IMPLIES P Q))
           (MY-CNF-DNF (FCONS-TERM* (QUOTE OR) (DUMB-NEGATE-LIT P) Q)
                    FLG))
          ((MATCH TERM (IFF P Q))
           (MY-CNF-DNF (FCONS-TERM* (QUOTE IF) P Q (FCONS-TERM* (QUOTE NOT) Q))
                       FLG))
          (T (LIST (LIST TERM))))))

(DEFUN MY-GENERALIZE2 (TERM-VAR-LST)
  ;; takes an alist instead of two lists, sets CL to be NIL, and doesn't negate the hyps
  (ITERATE FOR LIT
           IN (SCRUNCH (ITERATE FOR SUBTERM-VAR IN TERM-VAR-LST
				APPEND	;nconc would be OK but I prefer not to rely on GET-TYPES that way
				(GET-TYPES (CAR SUBTERM-VAR) NIL)))
           COLLECT (SUBLIS-EXPR1 TERM-VAR-LST LIT)))

(DEFUN MY-INDUCTION-CANDIDATES (HYPS CONC)
  (LET ((CL-SET (LIST (APPEND (ITERATE FOR HYP-FLG IN HYPS
				    WHEN (AND (EQ (CDR HYP-FLG) (QUOTE A))
					      (NOT (EQUAL (CAR HYP-FLG) TRUE)))
				      COLLECT (DUMB-NEGATE-LIT (CAR HYP-FLG)))
			      (LIST CONC))))
	TYPE-ALIST)		;*** One could worry about whether TYPE-ALIST is ok here
    (PICK-HIGH-SCORES
      (FAVOR-COMPLICATED-CANDIDATES
       (COMPUTE-VETOES
	(TRANSITIVE-CLOSURE
	 (REMOVE-UNCHANGING-VARS
	  (ITERATE FOR CL IN CL-SET
		   NCONC (ITERATE FOR LIT IN CL
				  NCONC
				  (GET-CANDS LIT)))
	  CL-SET)
	 (FUNCTION MERGE-CANDS)))))))

(DEFUN MY-INDUCTION-FORMULA-OF-CANDIDATE (CAND CL-SET)
  (ITERATE FOR CL IN
	    (IND-FORMULA (ACCESS CANDIDATE TESTS-AND-ALISTS-LST CAND)
			 (CONS (ACCESS CANDIDATE
				       INDUCTION-TERM
				       CAND)
			       (ACCESS CANDIDATE OTHER-TERMS
				       CAND))
			 CL-SET)
	COLLECT
	  (FLATTEN-ANTECEDENTS
	    (MAKE-FORMULA-FROM-CLAUSE CL))))

(DEFUN MY-LEMMAS-USED-OF-CANDIDATE (CAND)
  (ACCESS JUSTIFICATION LEMMAS
	  (ACCESS CANDIDATE JUSTIFICATION CAND)))

(defun my-SIMPLIFY-CLAUSES-MAXIMALLY (cl-set)
  ;; I don't bother to set process-hist appropriately here 
  (iterate for cl in cl-set
	   with p-clauses
	   do
	   (progn (simplify-clause-maximally cl)
		  (setq p-clauses (union-equal (make-set process-clauses) p-clauses)))
	   finally (setq process-clauses p-clauses)))

(DEFUN MY-SUBLIS-VAR (ALIST FORM)
  ;; like SUBLIS-VAR but assuming that FORM is a term so that we can optimize for NIL,
  ;; which is a good idea where it's used, namely in MATCH-FREE-VARS1, since the original
  ;; UNIFY-SUBST there is generated from the user's dollar-variable substitution, which
  ;; will often be NIL, I think.
  (IF (NULL ALIST)
      FORM
      (SUBLIS-VAR ALIST FORM)))

(DEFUN MY-SUBST-EXPR (NEW OLD FORM)
  (IF (EQUAL OLD FORM)
      NEW
      (SUBST-EXPR NEW OLD FORM)))

(DEFUN MY-TYPE-ALIST-CLAUSE (CL)
; Same as old version of type-alist-clause, i.e. using ASSUME-TRUE-FALSE
; rather than SMART-ASSUME-TRUE-FALSE.
  (LET ((TYPE-ALIST TYPE-ALIST))
    (ITERATE FOR LIT IN CL WHILE (NOT (EQ TYPE-ALIST (QUOTE CONTRADICTION))) DO
	  (PROGN (ASSUME-TRUE-FALSE LIT)
		 (COND (MUST-BE-TRUE (SETQ TYPE-ALIST (QUOTE CONTRADICTION)))
		       (MUST-BE-FALSE NIL)
		       (T (SETQ TYPE-ALIST FALSE-TYPE-ALIST)))))
    TYPE-ALIST))

(DEFUN MY-TYPE-SET (TERM)
  (LET (X Y)
    (COND
     ((MATCH TERM (EQUAL X Y))
      (IF (EQUAL X Y)
	  TYPE-SET-TRUE
	  (IF (AND 
	       (QUOTEP X)
	       (QUOTEP Y))
	      TYPE-SET-FALSE
	      (TYPE-SET TERM))))
     ((MATCH TERM (NOT (EQUAL X Y)))
      (IF (EQUAL X Y)
	  TYPE-SET-FALSE
	  (IF (AND 
	       (QUOTEP X)
	       (QUOTEP Y))
	      TYPE-SET-TRUE
	      (TYPE-SET TERM))))
     (T (TYPE-SET TERM)))))

(DEFUN MY-UNION-EQUAL (L M)
; The following is a bit less efficient that UNION-EQUAL, but it
; deletes the duplicated elements from the second list rather
; than from the first.
  (APPEND L (ITERATE FOR X IN M UNLESS (MEMBER-EQUAL X L) COLLECT X)))

(DEFUN my-WRAPUP (WON-FLG)
  ;; doesn't print the "Q.E.D"
  (COND ((NOT (EQ LEMMA-STACK ORIG-LEMMA-STACK))
         (ITERPRI T)
         (ER HARD NIL WRAPUP |found| |a| |non-trivial| LEMMA-STACK EXCL)))
  (COND ((NOT (EQ LINEARIZE-ASSUMPTIONS-STACK
                  ORIG-LINEARIZE-ASSUMPTIONS-STACK))
         (ITERPRI T)
         (ER HARD NIL WRAPUP |found| |a| |non-trivial|
             LINEARIZE-ASSUMPTIONS-STACK EXCL)))
  ;;(IO (QUOTE FINISHED)
  ;;    NIL NIL NIL (LIST WON-FLG))
  (iterpri nil)
  (THROW (QUOTE PROVE)
         (COND (WON-FLG (QUOTE PROVED))
               (T NIL))))

(DEFUN NEW-CITIZEN-NAMES (C OLD-C)
  (ITERATE FOR TAIL ON C
	   WITH OLD-TOP-NAME = (CAR OLD-C) AND TEMP 
	   DO
	   (COND
	    ((EQ (CAR TAIL) OLD-TOP-NAME)
	     (IF (EQ TAIL OLD-C)
		 (RETURN (NREVERSE TEMP))
		 (RETURN (QUOTE *FAILED*))))
	    (T
	     (SETQ TEMP (APPEND (GET (CAR TAIL) 'SATELLITES)
				(CONS (CAR TAIL) 
				      TEMP)))))
	   ; if we never find OLD-TOP-NAME in C then we fail
	   FINALLY (RETURN (QUOTE *FAILED*))))

(DEFUN NEW-GOAL-NAMES (GOAL-NAME OTHER-GOALS N)
  (ITERATE FOR X FROM 1 TO N
	    AS I FROM (NEW-GOAL-SUFFIX
			GOAL-NAME
			(ITERATE FOR GL-INFO IN OTHER-GOALS
			      COLLECT (ACCESS GOAL GOAL-NAME (CAR GL-INFO))))
	COLLECT (CONS GOAL-NAME I)))

(DEFUN NEW-GOAL-SUFFIX (CURRENT-NAME GOAL-NAMES)
; returns the least positive integer n
; such that for every member of GOAL-NAMES of the form
; (CURRENT-NAME . m), with m>0, we have m<n.  It's important
; that such a "large" n is returned (instead of one which
; merely fills a hole) in case multiple new goals are
; created.
  (ITERATE FOR NAME IN GOAL-NAMES WITH SUFFIX = 0
	WHEN
	  (AND (CONSP NAME)
	       (EQUAL (CAR NAME) CURRENT-NAME)
	       (INTEGERP (CDR NAME))
	       (< 0 (CDR NAME)))
	  DO (SETQ SUFFIX
		   (IF (< SUFFIX (CDR NAME))
		       (CDR NAME)
		       SUFFIX))
	FINALLY (RETURN (1+ SUFFIX))))

(DEFUN NEW-PART-OF-STATE-STACK (STATE-STACK)
  (ITERATE FOR STATE IN STATE-STACK
	WITH FLG
	UNTIL FLG
	COLLECT
	  (PROGN
	    (COND ((MATCH (ACCESS STATE INSTRUCTION STATE) (START &))
		   (SETQ FLG T)))
	    STATE)))

(DEFUN NO-IO () NIL)

(DEFUN ONE-WAY-UNIFY11-$ (TERM1 TERM2)
; This treats all variables as constants except when the first character
; is $.  Like ONE-WAY-UNIFY11 (which it plagiarizes from), it mangles the
; global UNIFY-SUBST even when it doesn't succeed, and it does not initialize
; UNIFY-SUBST either but rather tries to extend it.
; The only changes from the code for ONE-WAY-UNIFY11 are in the VARIABLEP
; case and in the recursive calls being to ONE-WAY-UNIFY11-$ instead of
; to ONE-WAY-UNIFY11.
  (COND ((VARIABLEP TERM1)
	 (COND ((EQUAL (CHAR (STRING TERM1) 0) #\$)
		(COND
		  ((SETQ TEMP-TEMP (ASSOC-EQ TERM1 UNIFY-SUBST))
		   (EQUAL (CDR TEMP-TEMP) TERM2))
		  (T
		   (SETQ UNIFY-SUBST (CONS (CONS TERM1 TERM2)
					   UNIFY-SUBST)))))
	       (T
		(EQ TERM1 TERM2))))
	((FQUOTEP TERM1)
	 
;   Since TERM1 is the only one whose variables we instantiate, and is
;   constant, and all terms are in the QUOTE-normal form discussed in
;   CONS-TERM, these two terms unify iff they are EQUAL.
	 
	 (EQUAL TERM1 TERM2))
	((VARIABLEP TERM2) NIL)
	((EQ (FFN-SYMB TERM1) (FN-SYMB TERM2))
	 (COND ((EQ (FFN-SYMB TERM1) (QUOTE EQUAL))
		(LET ((SAVED-UNIFY-SUBST UNIFY-SUBST))
		  (COND ((AND (ONE-WAY-UNIFY11-$ (FARGN TERM1 1) (FARGN TERM2 1))
			      (ONE-WAY-UNIFY11-$ (FARGN TERM1 2)
						 (FARGN TERM2 2)))
			 T)
			(T (SETQ UNIFY-SUBST SAVED-UNIFY-SUBST)
			   (AND (ONE-WAY-UNIFY11-$ (FARGN TERM1 2)
						   (FARGN TERM2 1))
				(ONE-WAY-UNIFY11-$ (FARGN TERM1 1)
						   (FARGN TERM2 2))
				(SETQ COMMUTED-EQUALITY-FLG T))))))
	       (T (ITERATE FOR ARG1 IN (FARGS TERM1) AS ARG2 IN (SARGS TERM2)
			ALWAYS (ONE-WAY-UNIFY11-$ ARG1 ARG2)))))
	(T NIL)))

(defun pop-clause-set-completely ()
  (setq bash-clauses nil)
  (setq stack (remove-being-proved-from-stack stack))
  (loop
   (if (null stack)
       (my-wrapup t)
     (setq bash-clauses
	   (append bash-clauses
		   (pop-clause-set))))
   (setq stack (remove-being-proved-from-stack stack))))

(DEFUN PPRIND-ABBREVIATIONS (FMLA LEFTMARGIN RPARCNT PPRFILE ABBREVIATIONS)
  (PPRIND (UNTRANSLATE-ABB FMLA ABBREVIATIONS)
	  LEFTMARGIN RPARCNT PPRFILE))

(DEFUN PPRIND-WEAK? (FMLA LEFTMARGIN RPARCNT PPRFILE ABBREVIATIONS)
  (PPRIND (UNTRANSLATE-ABB-WEAK FMLA ABBREVIATIONS)
	  LEFTMARGIN RPARCNT PPRFILE))

(DEFUN PREPARE-FOR-COPY (V)
  (LET* ((EXPLODED-R (REVERSE (COERCE (STRING V) 'LIST)))
	 (SUFFIX (ITERATE FOR X IN EXPLODED-R
			  UNTIL (EQUAL X #\$)
			  COLLECT X)))
	(COND
	 ((NULL SUFFIX)
	  (CONS V 0))
	 ((ITERATE FOR X IN SUFFIX ALWAYS (DIGIT-CHAR-P X))
	  (CONS (PACK (REVERSE (NTHCDR (LENGTH SUFFIX) EXPLODED-R)))
		(1+ (READ-FROM-STRING (COERCE (REVERSE SUFFIX) 'STRING)))))
	 (T (CONS (PACK (LIST V '$)) 0)))))

(DEFUN PRIMITIVE-FREWRITE-RULEP (FREWRITE-RULE)
  ;; A fast (i.e. almost primitive) rewrite rule is PRIMITIVE if it contains
  ;; no function symbol other than ones which (1) are added by BOOT-STRAP and
  ;; have no definition, e.g. CAR, CDR, CONS, or else (2) are added by the shell
  ;; principle.
  
  (AND (ITERATE FOR FN IN (ALL-FNNAMES (NTH 3 FREWRITE-RULE))
		WITH EV
		ALWAYS
		(OR (AND (EQ (GET FN (QUOTE MAIN-EVENT)) (QUOTE GROUND-ZERO))
			 (NOT (GET FN (QUOTE SDEFN))))
		    (AND (SETQ EV (GET FN (QUOTE EVENT)))
			 (EQ (CAR EV) (QUOTE ADD-SHELL)))
		    (AND (GET FN (QUOTE MAIN-EVENT))
			 (SETQ EV (GET (GET FN (QUOTE MAIN-EVENT)) (QUOTE EVENT)))
			 (EQ (CAR EV) (QUOTE ADD-SHELL)))))
       (ITERATE FOR FN IN (ALL-FNNAMES (NTH 4 FREWRITE-RULE))
		WITH EV
		ALWAYS
		(OR (AND (EQ (GET FN (QUOTE MAIN-EVENT)) (QUOTE GROUND-ZERO))
			 (NOT (GET FN (QUOTE SDEFN))))
		    (AND (SETQ EV (GET FN (QUOTE EVENT)))
			 (EQ (CAR EV) (QUOTE ADD-SHELL)))
		    (AND (GET FN (QUOTE MAIN-EVENT))
			 (SETQ EV (GET (GET FN (QUOTE MAIN-EVENT)) (QUOTE EVENT)))
			 (EQ (CAR EV) (QUOTE ADD-SHELL)))))))

(DEFUN PRINT-GOAL-NAMES-AS-TREE (TOP-GOAL GOALS INDENTATION)
  ;; assumes that TOP-GOAL (which is a goal) is at the top of the tree
  (let ((gname (ACCESS GOAL GOAL-NAME TOP-GOAL))
	(*print-level* 3)
	(*print-length* 3))
    (ispaces indentation prove-file)
    (pprind
     (if (equal (access goal conc top-goal) true)
	 (pack (list '* gname))
       (pack (list '| | gname)))
     indentation 0 prove-file)
    (let* ((state-position (goal-creation-info gname))
	   (ind (1+ (+ indentation (length (princ-to-string gname)))))
	   (info (pack (list '| <<| (cdr state-position) '|>> |))))
      (boyer-moore-print info
			 ind)
      (boyer-moore-print (access state instruction (car state-position))
			 (+ ind (length (princ-to-string info)))) 
      (iterpri prove-file))
    (ITERATE FOR GL IN GOALS
	     WITH IMMEDIATE-DEPENDENT-NAMES =
	     (GOAL-IMMEDIATE-DEPENDENTS TOP-GOAL GOALS)
	     WHEN (MEMBER-EQUAL (ACCESS GOAL GOAL-NAME GL)
				IMMEDIATE-DEPENDENT-NAMES)
	     DO (PRINT-GOAL-NAMES-AS-TREE GL GOALS (+ INDENTATION 3)))))

(DEFUN PRINT-INTERACTIVE-PROOF-FAILURE (NAME)
  ;; returns NIL
  (BLURB (NAME) CR CR
	 |*******************************************************|
	 CR FAILED TO COMPLETE INTERACTIVE PROOF OF (!PPR NAME NIL) CR
	 |*******************************************************| CR CR))

(DEFUN PROVE-cl (FORM return-clauses-flg)
  ;; adapted from nqthm definition of PROVE
  (LET ((PROOF-START-TIME-IN-60THS (TIME-IN-60THS))
	 clauses-popped-p)
       (prog1
	   (let ((won
		  (CATCH (QUOTE PROVE)
			 (PROG (THM CLAUSES VARS)
			       (CHK-INIT)
			       (SETQ THM (TRANSLATE FORM))
			       (SETQ CLAUSES (PREPROCESS THM))
			       (SETUP FORM CLAUSES ABBREVIATIONS-USED)
			       LOOP(SETQ VARS (ITERATE FOR CL IN CLAUSES WITH ITERATE-ANS
						       DO (SETQ ITERATE-ANS
								(UNION-EQ (ALL-VARS-LST CL)
									  ITERATE-ANS))
						       FINALLY (RETURN ITERATE-ANS)))
			       (SETQ ELIM-VARIABLE-NAMES1 (SET-DIFF ELIM-VARIABLE-NAMES VARS))
			       (SETQ GEN-VARIABLE-NAMES1 (SET-DIFF GEN-VARIABLE-NAMES VARS))
			       (SIMPLIFY-LOOP CLAUSES)
			       ;; Here's the change from PROVE
			       (if return-clauses-flg
				   (progn (setq clauses-popped-p t)
					  (pop-clause-set-completely))
				 (SETQ CLAUSES (INDUCT (POP-CLAUSE-SET))))
			       (GO LOOP)))))
	     (when (and (not won)
			return-clauses-flg
			(not clauses-popped-p))
		   (setq bash-clauses (list nil)))
	     won)
	 (SETQ PROVE-TIME (- (- (TIME-IN-60THS) PROOF-START-TIME-IN-60THS)
			     IO-TIME)))))

(DEFUN PROVE1 (TERM return-clauses-flg)
  ;; we should be sure that HINTS is always bound when this function is called
 (let ((IN-PROVE-LEMMA-FLG T))
  (BLURB NIL CR |*****| |Now| |entering| |the| |theorem| |prover| |*****| |:| CR)
  (UNWIND-PROTECT
   (LET (NEW-PROVE-TIME NEW-IO-TIME PROVE-ANS)
	(LET ((PROVE-TIME 0)
	      (IO-TIME 0))
	     (SETQ PROVE-ANS (PROVE-cl TERM return-clauses-flg))
	     (SETQ NEW-PROVE-TIME PROVE-TIME)
	     (SETQ NEW-IO-TIME IO-TIME))
	(SETQ PROVE-TIME (+ PROVE-TIME NEW-PROVE-TIME))
	(SETQ IO-TIME (+ IO-TIME NEW-IO-TIME))
	PROVE-ANS)
   (PROGN (ITERATE FOR X IN HINT-VARIABLE-ALIST
		   DO (SET (CADR X) (CADDDR X)))
	  (SETQ LOCAL-DISABLEDP-HASH-FLAG NIL)))))

(defun push-goal-to-end (goal-name state)
  ;; assumes that goal-name is among other-goals
  (make-new-record
   (quote state)
   (list (cons (quote other-goals)
	       (iterate for gl-info in (access state other-goals state)
			with last-goal-info and temp
			when (or last-goal-info
				 (not (and (equal (access goal goal-name (car gl-info))
						  goal-name)
					   (setq last-goal-info gl-info))))
			collect gl-info into temp
			finally (return (nconc temp (list last-goal-info))))))
   state))

(DEFUN QUICK-OPENER-TYPE (VAR TERM)
; returns the constructor function symbol corresponding to
; the destructor DEST if TERM is (DEST VAR), and NIL otherwise
  (COND ((VARIABLEP TERM) NIL)
	((FQUOTEP TERM) NIL)
	((AND (ITERATE FOR PAIR IN SHELL-POCKETS
		    THEREIS
		    (MEMBER-EQ (FFN-SYMB TERM) (CDR PAIR)))
	      (EQ VAR (FARGN TERM 1)))
	 (ITERATE FOR PAIR IN SHELL-POCKETS
	       WHEN (MEMBER-EQ (FFN-SYMB TERM) (CDR PAIR))
	       DO (RETURN (CAR PAIR))))
	(T NIL)))

(DEFUN READ-INSTR (INSTR)
  ;; Returns NIL if the instruction isn't valid 
  (COND
   ((INTEGERP INSTR)
    (COND
     ((<= INSTR 0)
      (BLURB (INSTR) |**NO| |CHANGE**| |--| |Integer| |instructions| |must| |be| |positive| |,|
	  |but| (!PPR INSTR NIL) |is| |not| EXCL CR))
     (T (LIST (QUOTE DIVE) INSTR))))
   ((MACRO-COMMAND-DEF (INSTR-NAME INSTR))
    (COND ((ATOM INSTR)
	   (SETQ INSTR (LIST INSTR))))
    (COND ((CHK-ACCEPTABLE-MACRO-COMMAND (CAR INSTR) (LENGTH (CDR INSTR)))
	   (READ-INSTR (APPLY
			(PACK (LIST *PC-MACRO*-PREFIX (CAR INSTR)))
			(EVAL-INSTR-ARGS (CDR INSTR)))))
	  (T (BLURB (INSTR) |**NO| |CHANGE**| |--| |Inappropriate| |number| |of| |arguments|
		 |for| |the| |macro| |command| (!PPR INSTR (QUOTE |.|)) CR))))
   ((CHK-ACCEPTABLE-INSTR INSTR)
    (COND
     ((ATOM INSTR)
      INSTR)
     ((NULL (CDR INSTR))
      (CAR INSTR))
     (T
      (CONS (CAR INSTR)
	    (EVAL-INSTR-ARGS (CDR INSTR))))))))

(DEFUN READ-INSTR-for-prove-lemma (INSTR)
  ;; Returns NIL if the instruction isn't valid 
  (COND
   ((INTEGERP INSTR)
    (COND
     ((<= INSTR 0)
      (BLURB (INSTR) |**NO| |CHANGE**| |--| |Integer| |instructions| |must| |be| |positive| |,|
	     |but| (!PPR INSTR NIL) |is| |not| EXCL CR))
     (T (LIST (QUOTE DIVE) INSTR))))
   ((MACRO-COMMAND-DEF (INSTR-NAME INSTR))
    (BLURB (INSTR) |**NO| |CHANGE**| |--| |Macro| |commands| |such| |as|
	   (!ppr instr nil) |are| |not|
	   |allowed| |during| |noninteractive| |replay| |.| CR))
   ((CHK-ACCEPTABLE-INSTR INSTR)
    (COND
     ((ATOM INSTR)
      INSTR)
     ((NULL (CDR INSTR))
      (CAR INSTR))
     (T instr)))))

(defun remove-being-proved-from-stack (stack)
  (cond ((null stack)
	 nil)
	((eq (caar stack)
	     (quote being-proved))
	 (remove-being-proved-from-stack (cdr stack)))
	(t stack)))

(DEFUN REWRITE-CURRENT-TERM (TERM LEMMA ID-IFF)
; Returns either NIL (when lemma doesn't apply) or else the
; LEMMA consed on to the substitution which one-way-unifies the
; left-hand-side of the lemma's conclusion with the given TERM.
; The lemma is returned in the form of an equality; if it can't
; be made an equality (relative to the given ID-IFF flag), then
; NIL is returned.
;    Finally: the lemma is checked for not being shelved.  If I
; want to remove this restiction in some particular situation I
; can of course rebind SHELVED-LEMMAS to NIL locally.
;    PERHAPS I should change the storage mechanism so that all lemmas
; are originally stored as equalities whenever possible.
  (COND
   ((AND (NOT (MEMBER-EQ (ACCESS REWRITE-RULE NAME LEMMA) SHELVED-LEMMAS))
	 (SETQ LEMMA (MAKE-EQUALITY LEMMA ID-IFF))
	 (ONE-WAY-UNIFY
	  (FARGN (ACCESS REWRITE-RULE CONCL LEMMA) 1)
	  TERM))
    (CONS LEMMA UNIFY-SUBST))))

(DEFUN ROSE-FUNCTIONP (FN-SYMB)
  (GET FN-SYMB 'TYPE-PRESCRIPTION-LST))

(DEFUN SAFECDR (X)
  (IF (CONSP X) (CDR X) NIL))

(DEFUN SAVED-STATE-STACK-LIST (STATE-STACK-LIST)
  ;; It's important for undoing to at least put something there to undo.
  (let (query-result)
    (cond
     (state-stack-list
      (blurb ((x (iterate for name-state-stack in state-stack-list collect (car name-state-stack))))
	     CR |**NOTE** --| |You| |have|
	     (PLURAL? X |some| |an|)
	     |unfinished| |interactive|
	     (PLURAL? X |proofs| |proof|)
	     |in| |your| |current| |library| |,| |namely| |for| |the|
	     (PLURAL? X |events| |event|)
	     (!LIST X) |.| |However| |,| |this| |should| |not| |be| |a| |problem| |.|)
      (cond
       (SAVE-STATE-STACK-LIST-QUERY-FLG
	(blurb nil |#| |Do| |you| |want| |to| |proceed| |to| |save| |state| |only|
	       |for| |incomplete| |partial| |proofs| |(P)| |,| |save| |state| |for|
	       |all| |partial| |proofs| |(S)| |,| |or| |save| |state| |for| |no|
	       |partial| |proofs| |(N)? |)
	(setq query-result
	      (iterate with ans = (read nil)
		       while (not (member-eq ans '(P S N)))
		       do (progn (blurb nil cr |Please| |answer| |P| |,| |S| |,| |or| |N| |: |)
				 (setq ans (read nil)))
		       finally (return ans))))
       (t (setq query-result 'S)))
      (blurb nil cr |proceeding| |with| |making| |library| |....| cr)
      (iterate for name-state-stack in state-stack-list
	       collect
	       (cond
		((or (eq query-result 's)
		     (and (eq query-result 'p)
			  (not (using-as-previous-event (car name-state-stack)))))
		 (cons (car name-state-stack)
		       (cons (car (cdr name-state-stack))
			     (our-last (cdr name-state-stack)))))
		(t (cons (car name-state-stack) nil))))))))

(DEFUN SCHEMA-MEASURE-RELATION-POSSIBILITIES (HYPS CONC)
  (ITERATE FOR CAND IN (MY-INDUCTION-CANDIDATES HYPS CONC)
	WITH CL-SET = (LIST (LIST (MAKE-IMPLICATION
				    (ASSUMPTIONS HYPS)
				    CONC)))
	COLLECT (GET-SCHEMA-MEASURE-RELATION
		  CAND
		  CL-SET)))

(DEFUN SHORT-HELP (INSTR-NAME)
  (EVAL (PACK (LIST INSTR-NAME (QUOTE -) (QUOTE SHORT-HELP-MESSAGE)))))

(DEFUN SHOW-FREE-VARIABLES-IN-LEMMA (LEMMA)
; The following assumes that the LEMMA is a rewrite rule of the form
; (EQUAL T1 T2), and returns LEMMA modified so that all of the free
; variables occurring in its hypotheses and conclusion are prefixed
; with dollar signs "$".  Here, a variable is  free  if it does not
; occur in T1 above (i.e. on the left side of the equality).  ACTUALLY,
; what is returned is the new LEMMA consed on to the list of free
; variables.
  (LET (FREE-VARS SUBST
	(HYPS (ACCESS REWRITE-RULE HYPS LEMMA))
	(CONCL (ACCESS REWRITE-RULE CONCL LEMMA)))
    (SETQ FREE-VARS
	  (SET-DIFF (UNION-EQ (ALL-VARS (FARGN CONCL 2))
			      (ALL-VARS-LST HYPS))
		    (ALL-VARS (FARGN CONCL 1))))
    (SETQ SUBST
	  (ITERATE FOR VAR IN FREE-VARS
		COLLECT
		(CONS VAR (PACK (LIST (QUOTE $) VAR)))))
    (CONS
      (MAKE-NEW-RECORD
	(QUOTE REWRITE-RULE)
	(LIST (CONS (QUOTE CONCL) (SUBLIS-VAR SUBST CONCL))
	      (CONS (QUOTE HYPS) (SUBLIS-VAR-LST SUBST HYPS)))
	LEMMA)
      FREE-VARS)))

(DEFUN SHOW-VARIABLES-WITH-GOALS (VARS STATE)
  (ITERATE FOR VAR IN VARS
	   WITH OCCUR-GOALS AND DEAD-VARS AND LIVE-FLG AND
	   GOAL-VARS-ALIST =
	   (ITERATE FOR GL IN (ALL-GOALS STATE)
		    COLLECT (CONS (ACCESS GOAL GOAL-NAME GL)
				  (ALL-VARS-GOAL GL))) 
	   DO (COND
	       ((SETQ OCCUR-GOALS
		      (ITERATE FOR GOALNAME-VARS IN GOAL-VARS-ALIST
			       WHEN (MEMBER-EQ VAR (CDR GOALNAME-VARS))
			       COLLECT (CAR GOALNAME-VARS)))
		(BLURB (VAR OCCUR-GOALS) | | (!PPR VAR NIL) |occurs| |in| |the|
		       (PLURAL? OCCUR-GOALS |goals| |goal|)
		       (!PPR-LIST OCCUR-GOALS (QUOTE |.|)) CR)
		(SETQ LIVE-FLG T))
	       (T
		(SETQ DEAD-VARS (CONS VAR DEAD-VARS))))
	   FINALLY
	   (PROGN
	    (WHEN (NOT LIVE-FLG)
		  (BLURB NIL |There| |are| |no| |free| |variables| |which|
			 |occur| |in| |any| |goals| |in| |the| |current|
			 |state| |.| CR))
	    (WHEN DEAD-VARS
		  (BLURB (LIVE-FLG DEAD-VARS
				   (COMM (CONS 'PUT (ITERATE FOR V IN DEAD-VARS
							     COLLECT (LIST V 0)))))
			 CR NOTE |:| (!PPR-LIST DEAD-VARS NIL)
			 (PLURAL? DEAD-VARS |do| |does|) |not| |occur|
			 |in| |any| |goals| |at|
			 |the| |moment| |.| |This| |is| |a| |harmless| |state| |of|
			 |affairs| |,| |though| |it| |could| |cause| |warning|
			 |messages| |in| INDUCT |and| GENERALIZE |commands| |.|
			 |If| |you| |really| |want| |to| |get| |rid| |of|
			 (PLURAL? DEAD-VARS
				  (PROGN |these| |free| |variables|)
				  (PROGN |this| |free| |variable|))
			 |,| |submit| |the| |command|
			 |(for| |example)| (!PPR COMM (QUOTE |.|)) CR)))))

(defun split-return-clauses-from-hints (hints)
  ;; removes any atom from hints and returns (cons ? real-hints) where
  ;; ? is that atom (or, NIL if there is none)
  (if (atom hints)
      (cons hints nil)
    (if (atom (car hints))
	hints
      (let ((hints-return (split-return-clauses-from-hints (cdr hints))))
	;; so, hints-return is of the form (cons ? real-hints)
	(cons (car hints-return)
	      (cons (car hints) (cdr hints-return)))))))

(defun start-term-for-interactive-prove-termination (formals rm machine)
  (dumb-conjoin
   (append
    (cond
     ((AND (EQ *1*THM-MODE$ (QUOTE NQTHM))
	   (EQ (CAR RM) (QUOTE ORD-LESSP)))
      (LIST (CONS-TERM (QUOTE ORDINALP)
		       (LIST (CADR RM))))))
    (ITERATE FOR X IN MACHINE
	     collect
	     (make-formula-from-clause
	      (NCONC1 (ITERATE
		       FOR H IN (ACCESS TESTS-AND-CASE TESTS X)
		       COLLECT (NEGATE-LIT H))
		      (CONS-TERM (CAR RM)
				 (LIST (SUB-PAIR-VAR
					FORMALS
					(ACCESS
					 TESTS-AND-CASE
					 CASE X)
					(CADR RM))
				       (CADR RM)))))))))

(DEFUN SUBLIS-VAR-GOAL (SUBST GOAL)
  ;; returns the result of substituting SUBST everywhere in the given goal
  (MAKE-NEW-RECORD
   (QUOTE GOAL)
   (LIST (CONS (QUOTE CONC)
	       (SUBLIS-VAR SUBST (ACCESS GOAL CONC GOAL)))
	 (CONS (QUOTE HYPS)
	       (SUBLIS-VAR-HYPS SUBST (ACCESS GOAL HYPS GOAL))))
   GOAL))

(DEFUN SUBLIS-VAR-HYPS (SUBST HYPS)
  (ITERATE FOR PAIR IN HYPS COLLECT
	   (CONS (SUBLIS-VAR SUBST (CAR PAIR))
		 (CDR PAIR))))

(DEFUN SUBSTP (SUB ABBREVIATIONS)
  (ITERATE FOR PAIR IN SUB
	ALWAYS
	(AND
	  (MATCH PAIR (LIST & &))
	  (VARIABLEP (TRANSLATE (CAR PAIR)))
	  (TRANSLATE-AND-CHK-ABB (CADR PAIR) ABBREVIATIONS))))

(DEFUN SWAP-GOALS (STATE NEW-NAME GOAL-AND-INFO)
; assumes that NEW-NAME is the name of goal in the CAR of GOAL-AND-INFO,
; which is not the current goal  -- returns a new state in which the
; given goal is the new goal and the current goal is put in with
; OTHER-GOALS
  (LET-FIELDS
    GOAL-NAME			;for the compiler's tranquility
    (MAKE-NEW-RECORD
      (QUOTE STATE)
      (LIST (CONS (QUOTE OTHER-GOALS)
		  (CONS (LIST GOAL CURRENT-TERM GOVERNORS CURRENT-ADDR-R)
			(ITERATE FOR GL-INFO IN OTHER-GOALS
			      WHEN (NOT
				     (EQUAL (ACCESS GOAL GOAL-NAME (CAR GL-INFO))
					    NEW-NAME))
				COLLECT GL-INFO)))
	    (CONS (QUOTE CURRENT-TERM)
		  (CADR GOAL-AND-INFO))
	    (CONS (QUOTE GOVERNORS) (CADDR GOAL-AND-INFO))
	    (CONS (QUOTE CURRENT-ADDR-R) (CADDDR GOAL-AND-INFO))
	    (CONS (QUOTE GOAL) (CAR GOAL-AND-INFO)))
      STATE)))

(DEFUN TAUTOLOGYP (TERM)
; The T flag below indicates that we are only checking whether TERM is
; a tautology "in the propositional sense".
  (EQUAL (NORMALIZE-IFS (EXPAND-BOOT-STRAP-NON-REC-FNS TERM) NIL NIL T)
	 TRUE))

(DEFUN TERM-ID-IFF (TERM ADDRESS ID-IFF)
; The property we want is that if one substitutes an equivalent subterm
; of TERM at the given address (equivalent modulo the flag returned by
; this function, that is), then the resulting term is equivalent modulo
; the ID-IFF argument to the original TERM.
  (COND
    ((NULL ADDRESS)
     ID-IFF)
    ((EQ (FN-SYMB TERM) (QUOTE IF))
     (IF (INT= (CAR ADDRESS) 1)
	 (TERM-ID-IFF (ARGN TERM 1) (CDR ADDRESS) (QUOTE IFF))
	 (TERM-ID-IFF (ARGN TERM (CAR ADDRESS)) (CDR ADDRESS) ID-IFF)))
    ((MEMBER-EQ (FN-SYMB TERM) (QUOTE (AND OR IMPLIES IFF NOT)))
     (TERM-ID-IFF (ARGN TERM (CAR ADDRESS)) (CDR ADDRESS) (QUOTE IFF)))
    (T
     (TERM-ID-IFF (ARGN TERM (CAR ADDRESS)) (CDR ADDRESS) (QUOTE ID)))))

(DEFUN TOGGLE-SHELVED-LEMMAS ()
  (SETQ SHELVED-LEMMAS-FLG
	(NOT SHELVED-LEMMAS-FLG)))

(DEFUN TOTALLY-ENABLEDP (FN FREWRITE-DISABLED-RULES)
  (AND
    (NOT (MEMBER-EQ FN FREWRITE-DISABLED-RULES))
    (OR
      (NOT (GET FN (QUOTE SDEFN)))
      (NOT (INTERSECTP FREWRITE-DISABLED-RULES
		       (ALL-FNNAMES (CADR (GET-DEFN FN))))))))

(DEFUN TRANSLATE-ABB (X ABBREVIATIONS)
  (LET (BADVARS)
    (COND
      ((SETQ BADVARS
	     (ITERATE FOR VAR IN (ALL-ATOMS X)
		   WITH AT-VARS = (ITERATE FOR PAIR IN ABBREVIATIONS
					COLLECT (CAR PAIR))
		   WHEN (AND (SYMBOLP VAR)
			     (EQUAL (CHAR (STRING VAR) 0) #\@)
			     (NOT (MEMBER-EQUAL VAR AT-VARS)))
		     COLLECT VAR))
       (ER SOFT (BADVARS) (!LIST BADVARS)
	   (PLURAL? BADVARS |are| |is|) |not| |in| |the|
	   |current| |abbreviations| |list| |.|))
      (T
       (TRANSLATE (SUBLIS ABBREVIATIONS X))))))

(DEFUN TRANSLATE-AND-CHK-ABB (X ABBREVIATIONS)

;   This function is equivalent to TRANSLATE-ABB but comments upon the
;   the translated text before returning it.

  (SETQ X (TRANSLATE-ABB X ABBREVIATIONS))
  (QUOTATIONS-CHK X)
  X)

(DEFUN TYPE-SET-HYPS (TERM NEWTERM)
; Returns a list of hypotheses to be added about NEWTERM based on the
; type set of TERM, with the AND structure flattened.  For example,
; (TYPE-SET-HYPS '(APPEND X '3) 'Y) evaluates to ((OR (LISTP Y) (NUMBERP Y))).
  (LET (TYPE-ALIST TYPE-SET)
    (SETQ TYPE-SET (TYPE-SET TERM))
    (COND
      ((EQUAL TYPE-SET TYPE-SET-UNKNOWN)
       NIL)
      ((> TYPE-SET 0)
       (LIST
	 (DUMB-DISJOIN
	   (ITERATE FOR REC-NUMBER IN RECOGNIZER-ALIST
		 WHEN (> (TSLOGAND (CDR REC-NUMBER) TYPE-SET) 0)
		   COLLECT (FCONS-TERM* (CAR REC-NUMBER) NEWTERM)))))
      (T
       (ITERATE FOR REC-NUMBER IN RECOGNIZER-ALIST
	     WHEN (INT= (TSLOGAND (CDR REC-NUMBER) TYPE-SET) 0)
	       COLLECT
		 (FCONS-TERM* (QUOTE NOT)
			      (FCONS-TERM* (CAR REC-NUMBER) NEWTERM)))))))

(DEFUN TYPE-SET-NON-DEDUCIBLE (TERMS TYPE-ALIST)
; Returns those terms in TERMS which do not follow from the ASSUMPTIONS
; implicit in TYPE-ALIST by type-set reasoning.
    (COND
     ((EQ TYPE-ALIST (QUOTE CONTRADICTION))
      NIL)
     (T
      (ITERATE FOR TERM IN TERMS
	       WHEN (TSLOGSUBSETP TYPE-SET-FALSE (MY-TYPE-SET TERM))
	       COLLECT TERM))))

(DEFUN UNDOLLAR-FY (TERM)
  (COND
    ((ATOM TERM)
     (LET ((EXPLODED (COERCE (STRING TERM) 'LIST)))
       (COND ((EQUAL (CAR EXPLODED) #\$)
	      (PACK (CDR EXPLODED)))
	     (T TERM))))
    ((FQUOTEP TERM)
     TERM)
    (T
     (FCONS-TERM (FFN-SYMB TERM)
		 (ITERATE FOR ARG IN (FARGS TERM)
		       COLLECT (UNDOLLAR-FY ARG))))))

(DEFUN UNDOLLAR-FY-LST (TERMS)
  (ITERATE FOR TERM IN TERMS COLLECT (UNDOLLAR-FY TERM)))

(DEFUN UNION-EQ-LIST (X)
  (IF (CONSP X)
      (IF (CONSP (CDR X))
	  (UNION-EQ (CAR X) (UNION-EQ-LIST (CDR X)))
	  (CAR X))
      NIL))

(DEFUN UNPROVED-DEPENDENTS (GL ALL-GOALS)
; returns the dependents of GL among ALL-GOALS which have
; non-TRUE conclusions
  (ITERATE FOR GOAL IN ALL-GOALS
	WITH DEPENDENTS = (ACCESS GOAL DEPENDS-ON GL)
	WHEN (AND (NOT (EQUAL (ACCESS GOAL CONC GOAL) TRUE))
		  (MEMBER-EQUAL (ACCESS GOAL GOAL-NAME GOAL)
			  DEPENDENTS))
	  COLLECT GOAL))

(DEFUN UNTRANSLATE-ABB (EVG ABBREVIATIONS)
  ;; used in definition of PPRIND-ABBREVIATIONS
  (UNTRANSLATE (ABBREVIATE EVG ABBREVIATIONS)))

(DEFUN UNTRANSLATE-ABB-WEAK (EVG ABBREVIATIONS)
  ;; used in definition of PPRIND-WEAK?
  (UNTRANSLATE-WEAK (ABBREVIATE EVG ABBREVIATIONS)))

(DEFUN UNTRANSLATE-WEAK (TERM)

; Copy the term X, introducing some notational abbreviations but
; without altering the structure of X with abbreviations that
; change the arity (e.g., (PLUS A (PLUS B C)) is returned rather
; than (PLUS A B C), and CADR does not appear in any result).
; We assume that X is already known to be an output of TRANSLATE.
  
  (COND ((VARIABLEP TERM) TERM)
	((QUOTEP TERM) (UNTRANSLATE-QUOTE (CADR TERM)))
	(T (CONS (FFN-SYMB TERM)
		 (ITERATE FOR ARG IN (FARGS TERM)
		       COLLECT (UNTRANSLATE-WEAK ARG))))))

(DEFUN UPDATE-COMMAND-BINDINGS (NEW OLD)
  (APPEND
    (ITERATE FOR X IN NEW
	  COLLECT
	  (IF (ATOM X)
	      (LIST X NIL)
	      (LIST (CAR X)
		    (EVAL-LOCAL (CADR X) OLD))))
    OLD))

(DEFUN UPDATE-FN-RULES-ALIST (RULES NEW-RULES)
  ;; Here RULES and NEW-RULES are function-rules alists
  (NCONC
   (ITERATE FOR FN-RULES IN NEW-RULES
	    WHEN (NOT (ASSOC-EQ (CAR FN-RULES) RULES))
	    COLLECT FN-RULES)
   (ITERATE FOR FN-RULES IN RULES
	    WITH TEMP	   
	    COLLECT 
	    (COND
	     ((SETQ TEMP (ASSOC-EQ (CAR FN-RULES) NEW-RULES))
	      (APPEND TEMP (CDR FN-RULES)))
	     (T FN-RULES)))))

(DEFUN UPDATE-OTHER-GOALS
       (OTHER-GOALS GOAL-NAME NEWCONC-AND-HYPS-LST NEW-GOAL-NAMES)
; Takes the OTHER-GOALS field of a state together with a GOAL-NAME and
; a list NEWCONC-AND-HYPS-LST of (CONS CONC HYPS) pairs for which a
; new subgoal is to be created with conclusion CONC and (active) hypotheses
; HYPS, and returns the revised OTHER-GOALS which: (1) includes these new
; goals (each depending on nothing) and (2) has each new goal name added to
; the DEPENDS-ON field of each existing goal in OTHER-GOALS that already
; includes GOAL-NAME in that field.  Also, a message about the creation of the
; new goals is sent to the screen.  Here, NEW-GOAL-NAMES is the list of names
; of the new goals.
  (LET (NEW-GOALS-WITH-INFO)
    (SETQ NEW-GOALS-WITH-INFO
	  (ITERATE FOR NEWCONC-AND-HYPS IN NEWCONC-AND-HYPS-LST
		AS NEWNAME IN NEW-GOAL-NAMES
		COLLECT
		  (LIST
		    (MAKE GOAL
			  (CAR NEWCONC-AND-HYPS)
			  (MAKE-HYP-FLGS (CDR NEWCONC-AND-HYPS))
			  NIL NEWNAME)
		    (CAR NEWCONC-AND-HYPS) NIL NIL)))
    (BLURB (NEW-GOAL-NAMES (N (LENGTH NEW-GOAL-NAMES))) CR
	   |Creating| (!PPR N NIL) |new| (PLURAL? N |subgoals| |subgoal|) |,|
	   (!PPR-LIST NEW-GOAL-NAMES (QUOTE |.|)) CR)
    (APPEND
      NEW-GOALS-WITH-INFO
      (CLOSE-ONCE NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS))))

(defun using-as-previous-event (event-name)
  ; a predicate indicating whether event-name is part of a previous-event hint
  ; of a later event.  If non-NIL, returns that later event.
  (iterate for name in chronology
	   with hints
	   when (match (get name 'event)
		       (prove-lemma & & & hints))
	   do
	   (cond
	    ((eq name event-name)
	     (return nil))
	    ((member-equal (list 'previous-event event-name) hints)
	     (return name)))))

;;************************************************************;;
;;                     THE FAST REWRITER                      ;;
;;************************************************************;;

; SPEC FOR THE FAST REWRITER/SIMPLIFIER:

; Every subterm should be simplified, where "simplified" means that
; no simplification rules apply.  The simplification rules are those
; given by the argument FREWRITE-RULES which do not have names
; occurring in the argument FREWRITE-DISABLED-RULES, together with a
; number of basic facts.  Here are those additional "basic facts".
; Note that we leave the specification for TYPE-SET incomplete -- but
; let's at least assume that TYPE-SET gives a type set which can
; always be proved to contain the actual type set of the given term
; (in every instance).  (I may want to be more explicit about
; TYPE-SET, but that may be a bit of work since it calls
; TYPE-PRESCRIPTION and ASSUME-TRUE-FALSE.)  Finally, the global
; variable MY-ALL-LEMMAS-USED should be updated appropriately.

; (1) The first argument of IF matters only propositionally.
; (2) Defined functions applied to explicit value terms should be
;     applied to these terms (assuming that the clock doesn't run
;     out).  But there's an exception:  when either the (top)
;     function symbol or a function symbol in its definition is
;     disabled, then this rule need not apply.  If a time limit is
;     exceeded then application is inhibited as well.
; (3) Rules expressed in FREWRITE-TYPE-PRED:
;     (a) (EQUAL x y) = F for explicit value terms x, y
;     (b) (EQUAL x y) = F if one of x, y is a bottom object and the
;         other is the application of a shell constructor
;     (c) (EQUAL x T) = x if x is known to have a boolean type set;
;         similarly for (EQUAL T y).
;     (d) (EQUAL x (EQUAL y z)) =
;         (IF (EQUAL y z) (EQUAL x T) (IF x F T)) ,
;         and similarly for (EQUAL (EQUAL x y) z) .
;     (e) (r x) = T if r is a recognizer and x has the appropriate
;         type set
;     (f) (r x) = F if r is a recognizer and the type set of x is
;         known not to include the type of r.
; (4) Rules expressed in FREWRITE-SOLIDIFY, where the type set of a
;     term is computed using TYPE-SET with the TYPE-ALIST specified
;     as an argument to FAST-REWRITE (but closing it according to the
;     commutativity of equality):
;     (a) A term with a type set of {TRUEP} simplifies to T.
;     (b) A term with a type set of {FALSEP} simplifies to F.
;     (c) For propositional simplification (as for the test of an
;         IF), a term with a type set disjoint from {FALSEP}
;         simplifies to T.
; (5) Rules expressed in FREWRITE-IF:
;     (a) (IF (IF a b c) x y) = (IF a (IF b x y) (IF c x y)).
;     (b) (IF T x y) = x  (also note (1) above)
;     (c) (IF F x y) = y
;     (d) (IF x y y) = y
;     (e) (IF x x F) = x
;     (f) (IF x T F) = x  if x is boolean
; (6) Quick openers:  (G x1 ... xn) is not left unopened if (i) G is
;     nonrecursive and enabled or (ii) G is recursive, enabled, and
;     (a) for some i, the i-th formal of G is one which
;         is, by itself, a measured subset whose COUNT decreases in
;         each call, where every call destructs v with an accessor for
;         the same shell constructor const; and
;     (b) xi is a bottom object or has type set disjoint from the
;         singleton type set {const}

; The proof of correctness of FAST-REWRITE is actually a corollary of
; the correctness of the following stronger theorem.  Let's say that
; a term is completely simplified if none of the above rules apply to
; it or any of its subterms.  Then:

; FREWRITE, FREWRITE-IF, FREWRITE-SOLIDIFY, FREWRITE-WITH-LEMMAS, and
; FREWRITE-TYPE-PRED are all correct (i.e. they all return terms
; which are completely simplified (where TYPE-ALIST is thought of as
; being applied to result of substituting ALIST into TERM), and they
; all preserve meaning in the appropriate senses), subject to the
; following restricions.  The CARs of the ALIST argument are
; completely simplified, the arguments are all of the appropriate
; type (whatever that means -- e.g., the TYPE-ALIST is really an
; association list of term-number pairs), FREWRITE-IF is only called
; on a term whose first argument is completely simplified, the ALIST
; argument (to each of these functions) is closed under commutation
; of equalities, and FREWRITE-WITH-LEMMAS is called only on
; non-variable terms all of whose proper subterms are completely
; simplified and where none of the rules from FREWRITE-TYPE-PRED
; (group 3 above) apply at the top level. The proof is by induction
; on the length of the (Lisp) computation.  Thus a separate
; termination argument is required.  This shouldn't be too hard; it's
; nearly an induction on the length of the term, but one also needs
; to notice somehow that one doesn't loop because of
; FREWRITE-WITH-LEMMAS.  A useful lemma is that if all the subterms
; of a given term are simplified, then this remains true after
; REWRITE-TYPE-PRED is applied to the term.

;SO -- the thing to do is probably to keep around two lists,
;frewrite-rules and more-frewrite-rules.  These remain
;unchanged throughout the session, i.e. they are fixed
;by (initialize-frewrite) and only changed by invoking
;this function.  On the other hand, the function FAST-REWRITE
;(which I'll have to replace FREWRITE with throughout the
;code) is passed two lists:  a list of fast rewrite rules
;and a list of NAMES of fast rewrite rules and defined
;functions.  The latter are intended to be disabled ones,
;and are passed for efficiency reasons.  But maybe it would
;be faster just to pass the available fast rewrite rules
;(with the disabled ones deleted) so that one won't have
;to continually walk through a list of names.  This would
;also make the code a bit simpler -- all one needs is the
;available fast rewrite rules and the unavailable non-recursive
;function names.  On the other hand, if one has a huge library
;then it's a shame to have to walk through the entire list
;of potential fast rewrite rules just in order to disable a
;few, rather than having to repeatedly look through a short
;list of disabled rules.  So I guess both lists should be
;passed, but it's up to the programmer how much preprocessing
;to do before doing such passing, i.e. how much modification
;of the available list is appropriate.

; Note that the FREWRITE, FREWRITE-IF, FREWRITE-SOLIDIFY, and
; FREWRITE-WITH-LEMMAS always return a non-NIL value (and in fact,
; I believe that they always return a term).

; First, the top-level call to the fast rewriter is made via the following
; function.  Notice that the ALIST argument of FREWRITE is set to NIL, i.e.
; at the top level there is as yet no notion of simplified arguments.

(DEFUN FAST-REWRITE (TERM TYPE-ALIST ID-IFF FREWRITE-RULES FREWRITE-DISABLED-RULES N)
    (iterate with old = term and new
	     do
	     (progn
	       (setq new
		     (FREWRITE old NIL TYPE-ALIST
			       ID-IFF FREWRITE-RULES FREWRITE-DISABLED-RULES N))
	       (if (equal old new)
		   (return new)
		   (setq old new)))))

; In the remainder of the functions, it will be important to maintain the
; invariant that the TYPE-ALIST is fleshed-out, i.e. is closed under
; commutation of equalities.  We also assume that no call is made to any
; of these functions in which TYPE-ALIST is CONTRADICTION.

(DEFUN FREWRITE (TERM ALIST TYPE-ALIST ID-IFF FREWRITE-RULES FREWRITE-DISABLED-RULES N)
  (LET (T1 T2 T3)
    (COND ((VARIABLEP TERM)
	   (COND ((SETQ T1 (ASSOC-EQ TERM ALIST))
		  (CDR T1))
		 (T (FREWRITE-SOLIDIFY TERM TYPE-ALIST ID-IFF))))
	  ((FQUOTEP TERM)
	   (IF (EQ ID-IFF (QUOTE IFF))
	       (IF (EQUAL TERM FALSE)
		   TERM
		   TRUE)
	       TERM))
	  ((MATCH TERM (IF T1 T2 T3))
	   (FREWRITE-IF (FREWRITE T1 ALIST TYPE-ALIST (QUOTE IFF)
				  FREWRITE-RULES FREWRITE-DISABLED-RULES N)
			T2 T3 ALIST TYPE-ALIST ID-IFF
			FREWRITE-RULES FREWRITE-DISABLED-RULES N))
	  (T (LET ((ARGS (ITERATE FOR ARG IN (FARGS TERM)
			       COLLECT (FREWRITE ARG ALIST TYPE-ALIST
						 (if (member-eq (ffn-symb term)
								'(and or implies not iff))
						     (quote iff)
						   (QUOTE ID))
						 FREWRITE-RULES FREWRITE-DISABLED-RULES N)))
		   TEMP
		   (FN (GET (FFN-SYMB TERM) (QUOTE LISP-CODE)))
		   (REDUCE-TERM-CLOCK REDUCE-TERM-CLOCK))
	       (COND ((AND FN
			   (ITERATE FOR ARG IN ARGS ALWAYS (QUOTEP ARG))
			   (OR (NULL (GET (FFN-SYMB TERM) (QUOTE EVENT)))
			       (NULL (CDAR DEFINED-FUNCTIONS-TOGGLED)))
			   (TOTALLY-ENABLEDP (FFN-SYMB TERM) FREWRITE-DISABLED-RULES)
			   ; the above is so that a function that e.g. calls V&C$ will
			   ; be treated cautiously if V&C$ is disabled
			   (NOT (EQ (QUOTE *1*FAILED)
				    (SETQ TEMP
					  (CATCH (QUOTE REDUCE-TERM)
					    (APPLY FN
						   (ITERATE FOR ARG IN ARGS
							 COLLECT (CADR ARG))))))))
		      (SETQ MY-ALL-LEMMAS-USED
			    (ADD-TO-SET (FFN-SYMB TERM) MY-ALL-LEMMAS-USED))
		      (IF (EQ ID-IFF (QUOTE IFF))
			  (IF (EQUAL TEMP *1*F)
			      FALSE
			      TRUE)
			  (LIST (QUOTE QUOTE) TEMP)))
		     (T (FREWRITE-WITH-LEMMAS
			  (FCONS-TERM (FFN-SYMB TERM) ARGS)
			  TYPE-ALIST ID-IFF
			  FREWRITE-RULES FREWRITE-DISABLED-RULES N))))))))

(DEFUN FREWRITE-HYP (TERM ALIST TYPE-ALIST FREWRITE-RULES FREWRITE-DISABLED-RULES N)
  ;; NOTE:  No longer do we automatically call the fast rewriter on (RECOGNIZER VARIABLE) terms
  (COND
   ((INT= N 0)
    (FREWRITE-SOLIDIFY (SUBLIS-VAR ALIST TERM) TYPE-ALIST (QUOTE IFF)))
   (T
    (FREWRITE TERM ALIST TYPE-ALIST (QUOTE IFF) FREWRITE-RULES FREWRITE-DISABLED-RULES (1- N)))))

(DEFUN FREWRITE-IF (T1 T2 T3 ALIST TYPE-ALIST ID-IFF
		       FREWRITE-RULES FREWRITE-DISABLED-RULES N)
  ; Notice that the way this is written, one shouldn't expect to be able
  ; to use any TYPE-ALIST hypotheses about an IF statement, since
  ; FREWRITE-SOLIDIFY is not called in here until other attempts to
  ; simplify are exhausted.
  (LET (T11 T12 T13)
    (COND ((MATCH T1 (IF T11 T12 T13))
	   ; Thus T11 can be true or false, or we would have already
	   ; eliminated it.
	   ; Also, we know that T11 is not itself an
	   ; if-expression.
	   (ASSUME-TRUE-FALSE T11)
	   (LET ((NEW-FALSE-TYPE-ALIST (FLESH-OUT-TYPE-ALIST FALSE-TYPE-ALIST))
		 (NEW-TRUE-TYPE-ALIST (FLESH-OUT-TYPE-ALIST TRUE-TYPE-ALIST)))
	     (FREWRITE-IF-FINISH
	      T11
	      (FREWRITE-IF T12 T2 T3
			   alist
			   NEW-TRUE-TYPE-ALIST
			   ID-IFF
			   FREWRITE-RULES FREWRITE-DISABLED-RULES N)
	      (FREWRITE-IF T13 T2 T3
			   alist
			   NEW-FALSE-TYPE-ALIST
			   ID-IFF
			   FREWRITE-RULES FREWRITE-DISABLED-RULES N)
	      TYPE-ALIST ID-IFF)))
	  (T (assume-true-false t1)
	     (COND (MUST-BE-TRUE (FREWRITE T2 ALIST TYPE-ALIST ID-IFF	
					   FREWRITE-RULES FREWRITE-DISABLED-RULES N))
		   (MUST-BE-FALSE (FREWRITE T3 ALIST TYPE-ALIST ID-IFF	
					    FREWRITE-RULES FREWRITE-DISABLED-RULES N))
		   (T (SETQ T2
			    (LET (FALSE-TYPE-ALIST
				  (NEW-TRUE-TYPE-ALIST
				   (FLESH-OUT-TYPE-ALIST TRUE-TYPE-ALIST)))
			      (FREWRITE T2
					alist
					NEW-TRUE-TYPE-ALIST
					ID-IFF
					FREWRITE-RULES FREWRITE-DISABLED-RULES N)))
		      (SETQ T3
			    (LET ((NEW-FALSE-TYPE-ALIST
				   (FLESH-OUT-TYPE-ALIST FALSE-TYPE-ALIST)))
			      (FREWRITE T3
					alist
					NEW-FALSE-TYPE-ALIST
					ID-IFF
					FREWRITE-RULES FREWRITE-DISABLED-RULES N)))
		      (FREWRITE-IF-FINISH T1 T2 T3 TYPE-ALIST ID-IFF)))))))

(DEFUN FREWRITE-IF-FINISH (T1 T2 T3 TYPE-ALIST ID-IFF)
  (COND ((EQUAL T2 T3) T2)
	((AND (EQUAL T1 T2) (EQUAL T3 FALSE)) T2)
	((AND (EQUAL T2 TRUE)
	      (EQUAL T3 FALSE)
	      (OUR-BOOLEAN T1))
	 T1)
	(T (FREWRITE-SOLIDIFY (FCONS-TERM* (QUOTE IF) T1 T2 T3)
			      TYPE-ALIST ID-IFF))))

(DEFUN FREWRITE-SOLIDIFY (TERM TYPE-ALIST ID-IFF)
  (LET ((TYPE-SET (TYPE-SET TERM)))
    (COND ((EQ ID-IFF (QUOTE IFF))
	   (COND ((EQUAL TYPE-SET TYPE-SET-FALSE) FALSE)
		 ((TSLOGSUBSETP TYPE-SET-FALSE TYPE-SET) TERM)
		 (T TRUE)))
	  (T
	   (COND ((EQUAL TYPE-SET TYPE-SET-FALSE) FALSE)
		 ((EQUAL TYPE-SET TYPE-SET-TRUE) TRUE)
		 (T TERM))))))

(DEFUN FREWRITE-TYPE-PRED (TERM)
; This function returns a pair (CONS NEWTERM FLAG), where the
; FLAG is T if and only if TERM = NEWTERM.
  (LET (T1 T2 PAIR TYPE-SET)
    (COND ((MATCH TERM (EQUAL T1 T2))
	   (COND ((EQUAL T1 T2)
		  (CONS TRUE T))
		 ((NOT-IDENT T1 T2)
		  (CONS FALSE T))
		 ((AND (OUR-BOOLEAN T1) (EQUAL TRUE T2))
		  (CONS T1 T))
		 ((AND (OUR-BOOLEAN T2) (EQUAL TRUE T1))
		  (CONS T2 T))
		 ((MATCH T2 (EQUAL & &))
		  (CONS (FCONS-TERM* (QUOTE IF)
				     T2
				     (FCONS-TERM* (QUOTE EQUAL) T1 TRUE)
				     (FCONS-TERM* (QUOTE IF) T1 FALSE TRUE))
			T))
		 ((EQUAL T1 FALSE)
		  (CONS (FCONS-TERM* (QUOTE IF) T2 FALSE TRUE)
			T))
		 ((EQUAL T2 FALSE)
		  (CONS (FCONS-TERM* (QUOTE IF) T1 FALSE TRUE)
			T))
		 ((MATCH T1 (EQUAL & &))
		  (CONS (FCONS-TERM* (QUOTE IF)
				     T1
				     (FCONS-TERM* (QUOTE EQUAL)
						  T2 TRUE)
				     (FCONS-TERM* (QUOTE IF)
						  T2 FALSE TRUE))
			T))
		 (T (CONS TERM NIL))))
	  ((SETQ PAIR (ASSOC-EQ (FFN-SYMB TERM)
			    RECOGNIZER-ALIST))
	   (SETQ TYPE-SET (TYPE-SET (FARGN TERM 1)))
	   (COND ((TSLOGSUBSETP TYPE-SET (CDR PAIR))
		  (CONS TRUE T))
		 ((INT= 0 (TSLOGAND TYPE-SET (CDR PAIR)))
		  (CONS FALSE T))
		 (T (CONS TERM NIL))))
	  (T (CONS TERM NIL)))))

(DEFUN FREWRITE-WITH-LEMMAS (TERM TYPE-ALIST ID-IFF
			     FREWRITE-RULES FREWRITE-DISABLED-RULES N)
  (LET (REWRITTEN-TERM NAME HYPS LHS RHS QUICK-OPENER DEF TEMP)
    (SETQ TERM (FREWRITE-SOLIDIFY TERM TYPE-ALIST ID-IFF))
    (COND ((CDR (SETQ TEMP (FREWRITE-TYPE-PRED TERM)))
	   (CAR TEMP))
	  ((FQUOTEP TERM)	;;Used to check (VARIABLEP TERM), but the spec
				;;makes this unnecessary.
	   TERM)		;;This seems to be an unfortunate restriction,
				;;in case the reduce clock ran out in FREWRITE --
				;;however, it makes the code below faster, I guess
				;;(e.g. one can use FFN-SYMB instead of FN-SYMB)
	  ((ITERATE FOR RULE IN (GET-FREWRITE-RULES (FFN-SYMB TERM) FREWRITE-RULES)
		    with saved-unify-subst
		   THEREIS
		     (AND (MATCH RULE (FREWRITE-RULE NAME HYPS LHS RHS))
			  (NOT (GET-FREWRITE-DISABLEDP NAME FREWRITE-DISABLED-RULES))
			  (ONE-WAY-UNIFY LHS TERM)
			  (progn (setq saved-unify-subst unify-subst) t)
			  ;;note that all CDRs in UNIFY-SUBST are in fact completely
			  ;;simplified, since we have assumed that every proper
			  ;;subterm of TERM is simplified.
			  (ITERATE FOR HYP IN HYPS
				ALWAYS
				  (EQUAL (FREWRITE-HYP HYP saved-unify-subst TYPE-ALIST
						       FREWRITE-RULES FREWRITE-DISABLED-RULES N)
					 TRUE))
			  (PROGN (SETQ MY-ALL-LEMMAS-USED
				       (ADD-TO-SET NAME MY-ALL-LEMMAS-USED))
				 (SETQ REWRITTEN-TERM
				       (FREWRITE RHS saved-UNIFY-SUBST TYPE-ALIST ID-IFF ;;**notice change
						 FREWRITE-RULES FREWRITE-DISABLED-RULES N))
				 T)))
	   REWRITTEN-TERM)
	  ((AND FNS-ALLOWED-FOR-FAST-REWRITE-FLG
		(GET (FFN-SYMB TERM) (QUOTE SDEFN))
		(SETQ DEF (GET-DEFN (FFN-SYMB TERM)))
		(NOT (GET-FREWRITE-DISABLEDP (FFN-SYMB TERM) FREWRITE-DISABLED-RULES))
		;;(AND (SETQ QUICK-OPENER (GET-QUICK-OPENER (FFN-SYMB TERM)))   ;)
		;; the line above was as follows, which caused FACTORIAL to open up way too much:
		(OR (NULL (SETQ QUICK-OPENER (GET-QUICK-OPENER (FFN-SYMB TERM))))
		    (ITERATE FOR ARG IN (FARGS TERM) AS TOKEN IN (CDR QUICK-OPENER)
			  THEREIS
			    (AND TOKEN
				 (OR (AND
				       (NVARIABLEP ARG)
				;;(OR (EQ (FN-SYMB ARG) TOKEN)
				       (BTM-OBJECTP ARG))	;;)
				;;*** Notice that no longer do we open up e.g. (times 3 x)
				     (NOT (TSLOGSUBSETP
					    (TSLOGBIT (CDR (ASSOC-EQ TOKEN
							       SHELL-ALIST)))
					    (TYPE-SET ARG))))))))
	   (SETQ MY-ALL-LEMMAS-USED (ADD-TO-SET (FFN-SYMB TERM) MY-ALL-LEMMAS-USED))
	   (FREWRITE (CADR DEF)
		     (ITERATE FOR V IN (CAR DEF) AS ARG IN (FARGS TERM)
			   COLLECT (CONS V ARG))
		     TYPE-ALIST
		     ID-IFF
		     FREWRITE-RULES FREWRITE-DISABLED-RULES N))
	  (T TERM))))
